from .utils import ObjectLookupDict


class Chain(object):

    """A chain is a sequence of phrases for a single channel. Each phrase can be
    transposed by a number of semitones.
    """

    def __init__(self, song, index):
        self.song = song
        self.index = index

        self.transposes = self.song.song_data.chain_transposes[index]
        self.phrases = ObjectLookupDict(
            self.song.song_data.chain_phrases[index], self.song.phrases)



from .bread_spec import STEPS_PER_TABLE
from .vendor.six.moves import range

class TableFX(object):

    def __init__(self, params, table_index, fx_index):
        self._params = params
        self._table_index = table_index
        self._fx_index = fx_index

    @property
    def command(self):
        """the effect's command"""
        return self._params.fx[self._table_index][self._fx_index]

    @command.setter
    def command(self, value):
        self._params.fx[self._table_index][self._fx_index] = value

    @property
    def value(self):
        """the command's parameter"""
        return self._params.val[self._table_index][self._fx_index]

    @value.setter
    def value(self, val):
        self._params.val[self._table_index][self._fx_index] = val

    def __eq__(self, other):
        return hasattr(other, '_params') and self._params == other._params

class Table(object):

    """Each table is a sequence of transposes, commands, and amplitude
    changes that can be applied to any channel."""

    def __init__(self, song, index):
        self._song = song
        self._index = index

        self._fx1 = [TableFX(self._song.song_data.table_cmd1, self._index, i)
                     for i in range(STEPS_PER_TABLE)]
        self._fx2 = [TableFX(self._song.song_data.table_cmd2, self._index, i)
                     for i in range(STEPS_PER_TABLE)]

    def __eq__(self, other):
        return self._fx1 == other._fx1 and self._fx2 == other._fx2

    @property
    def song(self):
        """the table's parent Song"""
        return self._song

    @property
    def index(self):
        """the table's index within its parent song's table of macro tables"""
        return self._index

    @property
    def fx1(self):
        """a list of the table's first effects, represented as TableFX
        objects"""
        return self._fx1

    @property
    def fx2(self):
        """a list of the table's first effects, represented as TableFX
        objects"""
        return self._fx2

    @property
    def envelopes(self):
        """a list of the table's volume envelopes"""
        return self._song.song_data.table_envelopes[self._index]

    @property
    def transposes(self):
        """a list of the table's volume transposes"""
        return self._song.song_data.table_transposes[self._index]

    def export(self):
        export_struct = []

        for i in range(STEPS_PER_TABLE):
            export_struct.append({
                "volume": self.envelopes[i],
                "transpose": self.transposes[i],
                "cmd1": [self.fx1[i].command, self.fx1[i].value],
                "cmd2": [self.fx2[i].command, self.fx2[i].value]
            })

        return export_struct

    def import_lsdinst(self, lsdinst_struct):
        table_data = lsdinst_struct['table']

        for i, table_row in enumerate(table_data):
            self.envelopes[i] = table_row["volume"]
            self.transposes[i] = table_row["transpose"]
            self.fx1[i].command, self.fx1[i].value = table_row["cmd1"]
            self.fx2[i].command, self.fx2[i].value = table_row["cmd2"]



import os
from nose.tools import assert_equal

from .project import load_lsdsng

SCRIPT_DIR = os.path.abspath(os.path.dirname(__file__))


def test_read_clocks():
    proj = load_lsdsng(
        os.path.join(SCRIPT_DIR, 'test_data', 'UNTOLDST.lsdsng'))

    project_clock = proj.song.clock
    total_clock = proj.song.global_clock

    print(project_clock)
    print(total_clock)
    print((total_clock.checksum))

    assert_equal(5, project_clock.hours)
    assert_equal(47, project_clock.minutes)

    assert_equal(57, total_clock.days)
    assert_equal(1, total_clock.hours)
    assert_equal(11, total_clock.minutes)


def test_set_local_clock():
    proj = load_lsdsng(
        os.path.join(SCRIPT_DIR, 'test_data', 'UNTOLDST.lsdsng'))

    project_clock = proj.song.clock

    project_clock.hours = 2
    project_clock.minutes = 22

    assert_equal(2, proj.song.clock.hours)
    assert_equal(22, proj.song.clock.minutes)


def test_set_global_clock():
    proj = load_lsdsng(
        os.path.join(SCRIPT_DIR, 'test_data', 'UNTOLDST.lsdsng'))

    proj.song.global_clock.days = 5
    proj.song.global_clock.hours = 14
    proj.song.global_clock.minutes = 20

    assert_equal(5, proj.song.global_clock.days)
    assert_equal(14, proj.song.global_clock.hours)
    assert_equal(20, proj.song.global_clock.minutes)

    assert_equal(39, proj.song.global_clock.checksum)



from .instrument import Instrument
from .instrument_mixins import VibratoMixin


class KitInstrument(Instrument, VibratoMixin):

    def __init__(self, song, index):
        super(KitInstrument, self).__init__(song, index)

    @property
    def volume(self):
        """the kit's volume; 0 to 3"""
        return self.data.volume

    @volume.setter
    def volume(self, value):
        self.data.volume = value

    @property
    def kit_1(self):
        """the index of the first kit in LSDJ's kit list"""
        return self.data.kit_1

    @kit_1.setter
    def kit_1(self, value):
        self.data.kit_1 = value

    @property
    def loop_1(self):
        """loop sample in kit 1 and start playing from an offset"""
        return self.data.loop_1

    @loop_1.setter
    def loop_1(self, value):
        self.data.loop_1 = value

    @property
    def keep_attack_1(self):
        """loop sample in kit 1 and start playing from beginning"""
        return self.data.keep_attack_1

    @keep_attack_1.setter
    def keep_attack_1(self, value):
        self.data.keep_attack_1 = value

    @property
    def length_1(self):
        """the length of kit 1's sound (0 means 'always play the sample to
        the end' and is displayed as AUT in LSDJ)"""
        return self.data.length_1

    @length_1.setter
    def length_1(self, value):
        self.data.length_1 = value

    @property
    def offset_1(self):
        """ kit 1's loop start point (if loop_1 is True and keep_attack_1 is
        False)"""
        return self.data.offset_1

    @offset_1.setter
    def offset_1(self, value):
        self.data.offset_1 = value

    @property
    def kit_2(self):
        """the index of the second kit in LSDJ's kit list"""
        return self.data.kit_2

    @kit_2.setter
    def kit_2(self, value):
        self.data.kit_2 = value

    @property
    def loop_2(self):
        """loop sample in kit 2 and start playing from an offset"""
        return self.data.loop_2

    @loop_2.setter
    def loop_2(self, value):
        self.data.loop_2 = value

    @property
    def keep_attack_2(self):
        """loop sample in kit 2 and start playing from beginning"""
        return self.data.keep_attack_2

    @keep_attack_2.setter
    def keep_attack_2(self, value):
        self.data.keep_attack_2 = value

    @property
    def length_2(self):
        """the length of kit 2's sound (0 means 'always play the sample to
        the end' and is displayed as AUT in LSDJ)"""
        return self.data.length_2

    @length_2.setter
    def length_2(self, value):
        self.data.length_2 = value

    @property
    def offset_2(self):
        """ kit 2's loop start point (if loop_2 is True and keep_attack_2 is
        False)"""
        return self.data.offset_2

    @offset_2.setter
    def offset_2(self, value):
        self.data.offset_2 = value

    @property
    def pitch(self):
        """sample pitch shift (8-bit integer)"""
        return self.data.pitch

    @pitch.setter
    def pitch(self, value):
        self.data.pitch = value

    @property
    def dist_type(self):
        """algorithm used when two kits are mixed together; ``clip``, ``shape``,
        ``shap2`` or ``wrap``"""
        return self.data.dist_type

    @dist_type.setter
    def dist_type(self, value):
        self.data.dist_type = value

    @property
    def half_speed(self):
        """if true, play samples at half their normal speed"""
        return self.data.half_speed

    @half_speed.setter
    def half_speed(self, value):
        self.data.half_speed = value

    def import_lsdinst(self, lsdinst_struct):
        super(KitInstrument, self).import_lsdinst(lsdinst_struct)

        self.volume = lsdinst_struct['data']['volume']

        self.kit_1 = lsdinst_struct['data']['kit_1']
        self.loop_1 = lsdinst_struct['data']['loop_1']
        self.keep_attack_1 = lsdinst_struct['data']['keep_attack_1']
        self.length_1 = lsdinst_struct['data']['length_1']
        self.offset_1 = lsdinst_struct['data']['offset_1']

        self.kit_2 = lsdinst_struct['data']['kit_2']
        self.loop_2 = lsdinst_struct['data']['loop_2']
        self.keep_attack_2 = lsdinst_struct['data']['keep_attack_2']
        self.length_2 = lsdinst_struct['data']['length_2']
        self.offset_2 = lsdinst_struct['data']['offset_2']

        self.pitch = lsdinst_struct['data']['pitch']
        self.dist_type = lsdinst_struct['data']['dist_type']

        VibratoMixin.import_lsdinst(self, lsdinst_struct)



class Clock(object):

    def __init__(self, clock_data):
        self._clock_data = clock_data

    @property
    def hours(self):
        """The total number of hours on the clock."""
        return self._clock_data.hours

    @hours.setter
    def hours(self, hours):
        self._clock_data.hours = hours

    @property
    def minutes(self):
        """The total number of minutes on the clock."""
        return self._clock_data.minutes

    @minutes.setter
    def minutes(self, minutes):
        self._clock_data.minutes = minutes

    def __repr__(self):
        return "%d hours, %d minutes" % (self.hours, self.minutes)


class TotalClock(object):

    def __init__(self, clock_data):
        """Constructor.

        We skip checksum computation on initial load, since LSDJ seems to be
        just fine with the global clock checksum being incorrect.

        :param clock_data: raw clock data from the parent song

        """
        self._clock_data = clock_data

    def _update_checksum(self):
        self._clock_data.checksum = (
            self._clock_data.days
            + self._clock_data.hours
            + self._clock_data.minutes)

    @property
    def days(self):
        """The total number of days on the clock."""
        return self._clock_data.days

    @days.setter
    def days(self, days):
        self._clock_data.days = days
        self._update_checksum()

    @property
    def hours(self):
        """The total number of hours on the clock."""
        return self._clock_data.hours

    @hours.setter
    def hours(self, hours):
        self._clock_data.hours = hours
        self._update_checksum()

    @property
    def minutes(self):
        """The total number of minutes on the clock."""
        return self._clock_data.minutes

    @minutes.setter
    def minutes(self, minutes):
        self._clock_data.minutes = minutes
        self._update_checksum()

    @property
    def checksum(self):
        """the clock's checksum (days + hours + minutes)"""
        return self._clock_data.checksum

    def __repr__(self):
        return "%d days, %d hours, %d minutes" % \
            (self.days, self.hours, self.minutes)

    def __eq__(self, other):
        return type(self) == type(other) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self.__eq__(other)



import os
import sys
from nose.tools import assert_equal, with_setup

from .utils import temporary_file

sys.path.append(
    os.path.dirname(os.path.abspath(os.path.join(__file__, os.path.pardir))))

from . import savfile as savfile

SAV_IN = os.path.join(os.path.dirname(__file__), "test_data", "lsdj.sav")

def test_project_save_load():

    with temporary_file() as SAV_OUT:
        sav = savfile.SAVFile(SAV_IN)
        sav.save(SAV_OUT)

        new_sav = savfile.SAVFile(SAV_OUT)

        assert_equal(sav, new_sav)



import os
from nose.tools import assert_false, assert_true

SCRIPT_DIR = os.path.abspath(os.path.dirname(__file__))

from .project import load_lsdsng

def test_wave_synth_overwrite_locks():
    test_project = os.path.join(SCRIPT_DIR, 'test_data', 'UNTOLDST.lsdsng')

    project = load_lsdsng(test_project)
    project.song.synths[4].phase_type = 'resync2'
    assert_false(project.song.wave_synth_overwrite_locks[0xf - 4],
                 "Expected wave synth overwrite lock to be disabled "
                 "after modifying synth parameters")
    assert_false(project.song.synths[4].wave_synth_overwrite_lock)

    project.song.synths[4].waves[5][3] = 0xd
    assert_true(project.song.wave_synth_overwrite_locks[0xf - 4],
                "Expected wave synth overwrite lock to be enabled "
                "after modifying wave frames")
    assert_true(project.song.synths[4].wave_synth_overwrite_lock)



from .instrument import Instrument
from .instrument_mixins import SoundLengthMixin, EnvelopeMixin, SweepMixin

MIXINS = [SoundLengthMixin, EnvelopeMixin, SweepMixin]


class NoiseInstrument(Instrument, SoundLengthMixin, EnvelopeMixin, SweepMixin):

    def __init__(self, song, index):
        super(NoiseInstrument, self).__init__(song, index)

    @property
    def s_cmd(self):
        """``free`` or ``stable``. When free, altering noise shape with the
        S command can sometimes mute the sound. When stable, sound will never
        be muted by accident. My understanding is that this setting exists for
        backwards-compatibility of behavior in old LSDJ instruments"""
        return self.data.s_cmd

    @s_cmd.setter
    def s_cmd(self, value):
        self.data.s_cmd = value

    def import_lsdinst(self, struct_data):
        super(NoiseInstrument, self).import_lsdinst(struct_data)

        self.s_cmd = struct_data['data']['s_cmd']

        for mixin in MIXINS:
            mixin.import_lsdinst(self, struct_data)



import os
import sys
import json
from nose.tools import assert_equal, assert_list_equal
from .vendor.six.moves import range

SCRIPT_DIR = os.path.abspath(os.path.dirname(__file__))

sys.path.append(os.path.join(SCRIPT_DIR, os.path.pardir))

from . import blockutils as bl


def test_simple_read_write():
    data = [i % 10 for i in range(bl.BLOCK_SIZE * 5 + 17)]

    reader = bl.BlockReader()
    writer = bl.BlockWriter()
    factory = bl.BlockFactory()

    block_ids = writer.write(data, factory)

    blocks = {}

    for i in block_ids:
        blocks[i] = factory.blocks[i]

    recovered_data = reader.read(blocks)

    assert_list_equal(data, recovered_data)


def test_sample_file():
    reader = bl.BlockReader()
    writer = bl.BlockWriter()
    factory = bl.BlockFactory()

    sample_song_blocks = os.path.join(
        SCRIPT_DIR, "test_data", "sample_song_blocks.json")

    with open(sample_song_blocks, "r") as fp:
        song_block_data = json.load(fp)

    song_blocks = {}

    # Dummy block for file table header
    factory.new_block()

    for i, key in enumerate(sorted(map(int, list(song_block_data.keys())))):
        song_blocks[i + 1] = factory.new_block()
        song_blocks[i + 1].data = bytearray(song_block_data[str(key)])

    with open(os.path.join(SCRIPT_DIR,
                           "test_data", "sample_song_compressed.json"),
              "r") as fp:
        compressed = json.load(fp)

    assembled_from_blocks = reader.read(song_blocks)
    assert_equal(assembled_from_blocks, compressed)

    factory = bl.BlockFactory()
    factory.new_block()

    block_ids = writer.write(assembled_from_blocks, factory)

    assert_equal(len(block_ids), 10)

    block_map = dict(factory.blocks)
    del block_map[0]

    assembled_from_write = reader.read(block_map)

    assert_equal(assembled_from_write, compressed)



import json

from .utils import assert_index_sane

from .wave_instrument import WaveInstrument
from .pulse_instrument import PulseInstrument
from .kit_instrument import KitInstrument
from .noise_instrument import NoiseInstrument

class Instruments(object):
    instrumentClasses = {
        "pulse": PulseInstrument,
        "wave": WaveInstrument,
        "kit": KitInstrument,
        "noise": NoiseInstrument
    }

    def __init__(self, song):
        self.song = song
        self.alloc_table = song.song_data.instr_alloc_table
        self.access_objects = []

        for index in range(len(self.alloc_table)):
            instr_type = self.song.song_data.instruments[index].instrument_type

            self.access_objects.append(
                self.instrumentClasses[instr_type](song, index))

    def _set_instrument_type(self, index, instrument_type):
        assert instrument_type in Instruments.instrumentClasses, (
            "Invalid instrument type '%s'" % (instrument_type))

        assert_index_sane(index, len(self.song.song_data.instruments))

        current_access_object = self.access_objects[index]

        # If this instrument is of a different type than we're currently
        # accessing, we've got to make a new access object of the
        # appropriate type
        if (current_access_object is None or
                current_access_object.type != instrument_type):

            self.access_objects[index] = (
                self.instrumentClasses[instrument_type](self.song, index))

            self.access_objects[index].type = instrument_type

    def __getitem__(self, index):
        assert_index_sane(index, len(self.alloc_table))

        if not self.alloc_table[index]:
            return None

        return self.access_objects[index]

    def as_list(self):
        return self.access_objects

    def allocate(self, index, instrument_type):
        self.alloc_table[index] = True
        self._set_instrument_type(index, instrument_type)

    def import_from_file(self, index, filename):
        """Import this instrument's settings from the given file. Will
        automatically add the instrument's synth and table to the song's
        synths and tables if needed.

        Note that this may invalidate existing instrument accessor objects.

        :param index: the index into which to import

        :param filename: the file from which to load

        :raises ImportException: if importing failed, usually because the song
          doesn't have enough synth or table slots left for the instrument's
          synth or table
        """

        with open(filename, 'r') as fp:
            self._import_from_struct(index, json.load(fp))

    def _import_from_struct(self, index, lsdinst_struct):
        instr_type = lsdinst_struct['data']['instrument_type']

        self.allocate(index, instr_type)

        instrument = self.song.instruments[index]
        instrument.name = lsdinst_struct['name']

        # Make sure we've got enough room for the table if we need it
        if 'table' in lsdinst_struct:
            table_index = self.song.tables.next_free()

            if table_index is None:
                raise ImportException(
                    "No available table slot in which to store the "
                    "instrument's table data")

            self.song.tables.allocate(table_index)
            instrument.table = self.song.tables[table_index]

        instrument.import_lsdinst(lsdinst_struct)



from .utils import ObjectLookupDict

# Need to define a custom lookup dict so that index=255 is treated as unset and
# doesn't throw an exception
class InstrumentLookupDict(ObjectLookupDict):
    def __getitem__(self, index):
        if self.id_list[index] == 255:
            return None

        return super(InstrumentLookupDict, self).__getitem__(index)

class Phrase(object):

    """A phrase is a sequence of notes for a single channel.
    """

    def __init__(self, song, index):
        self._song = song
        self._index = index
        self._instruments = InstrumentLookupDict(
            self._song.song_data.phrase_instruments[index],
            self._song.instruments)

    @property
    def song(self):
        """a reference to the phrase's parent song"""
        return self._song

    @property
    def index(self):
        """the phrase's index within its parent song's phrase table"""
        return self._index

    @property
    def instruments(self):
        """a list of Instruments, None where no instrument is defined"""
        return self._instruments

    @property
    def notes(self):
        """a list of the phrase's notes, one byte per note"""
        return self.song.song_data.phrase_notes[self._index]

    @property
    def fx(self):
        """a list of the phrase's effects, one byte per effect"""
        return self.song.song_data.phrase_fx[self._index]

    @property
    def fx_val(self):
        """a list of the phrase's effect parameters, one byte per effect"""
        return self.song.song_data.phrase_fx_val[self._index]



from .utils import assert_index_sane
from . import bread_spec

from .synth import Synth
from .table import Table
from .phrase import Phrase
from .chain import Chain
from .speech_instrument import SpeechInstrument

from .instruments import Instruments

from .bread_spec import INSTRUMENT_TYPE_CODE
from .filepack import DEFAULT_INSTRUMENT

from .clock import Clock, TotalClock

from .exceptions import ImportException

from .vendor.six.moves import range

# Number of channels
NUM_CHANNELS = 4


class AllocTable(object):

    def __init__(self, song, alloc_table, object_class):
        self.alloc_table = alloc_table

        self.access_objects = []

        for index in range(len(alloc_table)):
            self.access_objects.append(object_class(song, index))

    def __getitem__(self, index):
        assert_index_sane(index, len(self.alloc_table))

        if not self.alloc_table[index]:
            return None

        return self.access_objects[index]

    def allocate(self, index):
        self.alloc_table[index] = True

    def __len__(self):
        return len(self.alloc_table)

    def as_list(self):
        l = []
        for i in range(len(self.alloc_table)):
            if not self.alloc_table[i]:
                l.append(None)
            else:
                l.append(self.access_objects[i])

        return l

    def next_free(self):
        for i, occupied in enumerate(self.alloc_table):
            if not occupied:
                return i
        return None

class Grooves(object):

    def __init__(self, song):
        self.song = song

    def __getitem__(self, index):
        assert_index_sane(index, len(self.song.song_data.grooves))

        return self.song.song_data.grooves[index]


class Sequence(object):
    PU1 = "pu1"
    PU2 = "pu2"
    WAV = "wav"
    NOI = "noi"

    NO_CHAIN = 0xff

    def __init__(self, song):
        self.song = song

    def __getitem__(self, index):
        assert_index_sane(index, len(self.song.song_data.song))
        raw_chain = self.song.song_data.song[index]

        chain_objs = {}

        for channel in [Sequence.PU1,
                        Sequence.PU2,
                        Sequence.WAV,
                        Sequence.NOI]:
            chain_number = getattr(raw_chain, channel)

            if chain_number != Sequence.NO_CHAIN:
                chain = self.song.chains[chain_number]
                chain_objs[channel] = chain

        return chain_objs

    def __setitem__(self, index, value_dict):
        assert_index_sane(index, len(self.song.song_data.song))

        for channel in value_dict:
            assert (channel in [Sequence.PU1, Sequence.PU2, Sequence.WAV,
                                Sequence.NOI]), \
                ("Channel '%d' is not a valid channel" % (channel))

            chain = value_dict[channel]
            chain_number = chain.index

            assert_index_sane(chain_number,
                              len(self.song.song_data.chain_alloc_table))

            assert self.song.song_data.chain_alloc_table[chain_number], (
                "Assigning a chain (%d) that has not been allocated" % (
                    chain_number))

            setattr(self.song.song_data.song[index], channel, chain_number)

    def __str__(self):
        output_str = ''

        def add_line(line):
            output_str += line + '\n'


        add_line("   PU1 PU2 WAV NOI")

        for i, row in enumerate(self.song.song_data.song):
            add_line("%02x" % (i), end=' ')

            for channel in ["pu1", "pu2", "wav", "noi"]:
                chain_number = getattr(row, channel)

                if chain_number == Sequence.NO_CHAIN:
                    add_line(" --", end=' ')
                else:
                    add_line(" %02x" % (getattr(row, channel)), end=' ')
            add_line("")

        return output_str

class Synths(object):

    def __init__(self, song):
        self.song = song
        self.access_objects = []

        for index in range(bread_spec.NUM_SYNTHS):
            self.access_objects.append(Synth(self.song, index))

    def __getitem__(self, index):
        assert_index_sane(index, bread_spec.NUM_SYNTHS)

        return self.access_objects[index]

    def as_list(self):
        return self.access_objects


class Song(object):

    """A song consists of a sequence of chains, one per channel.
    """

    def __init__(self, song_data):
        # Check checksums
        assert song_data.mem_init_flag_1 == b'rb'
        assert song_data.mem_init_flag_2 == b'rb'
        assert song_data.mem_init_flag_3 == b'rb'

        # Everything we do to the song or any of its components should update
        # the song data object, so that we can rely on bread's writer to write
        # it back out in the right format
        self.song_data = song_data

        self._grooves = Grooves(self)
        self._speech_instrument = SpeechInstrument(self)
        self._synths = Synths(self)

        self._instruments = Instruments(self)

        # Stitch together allocated tables
        self._tables = AllocTable(
            song=self,
            alloc_table=self.song_data.table_alloc_table,
            object_class=Table)

        # Stitch together allocated phrases
        self._phrases = AllocTable(
            song=self,
            alloc_table=self.song_data.phrase_alloc_table,
            object_class=Phrase)

        # Stitch together allocated chains
        self._chains = AllocTable(
            song=self,
            alloc_table=self.song_data.chain_alloc_table,
            object_class=Chain)

        self._sequence = Sequence(self)

    def __str__(self):
        return str(self.sequence)

    @property
    def instruments(self):
        """the song's instrument table, represented as a list of
        Instrument objects"""
        return self._instruments

    @property
    def phrases(self):
        """the song's phrase table, represented as a list of Phrase objects"""
        return self._phrases

    @property
    def chains(self):
        """the song's chain table, represented as a list of Chain objects"""
        return self._chains

    @property
    def grooves(self):
        """the song's groove table"""
        return self._grooves

    @property
    def speech_instrument(self):
        """the song's speech instrument settings, represented as a
        SpeechInstrument object"""
        return self._speech_instrument

    @property
    def synths(self):
        return self._synths

    @property
    def clock(self):
        """the amount of time LSDJ has been used since the last memory
        reset, represented as a Clock object"""
        return Clock(self.song_data.clock)

    @property
    def global_clock(self):
        """the amount of time LSDJ has been used total, represented
        as a Clock object"""
        return TotalClock(self.song_data.total_clock)

    @property
    def song_version(self):
        """the song's version number"""
        return self.song_data.version

    @song_version.setter
    def song_version(self, version):
        self.song_data.version = version

    @property
    def sequence(self):
        """the song's sequence, showing the order in which chains are
        played on each of the four channels"""
        return self._sequence

    @property
    def tables(self):
        """the song's table of macro tables, represented as Table objects"""
        return self._tables

# For fields with a one-to-one correspondence with song data, we'll
# programmatically insert properties to avoid repetition
for field, doc in [("tempo", "the song's tempo"),
                   ("tune_setting", None),
                   ("key_delay", "the delay before key repeat is activated "
                    "for Game Boy buttons"),
                   ("key_repeat", "the key repeat speed for Game Boy buttons"),
                   ("font", "the selected LSDJ font"),
                   ("sync_setting", 'LSDJ\'s sync setting; one of ``"off"``, '
                    '``"slave"``, ``"master"``, ``"midi"``, ``"nano"``, and '
                    '``"keyboard"``'),
                   ("colorset", "the selected LSDJ colorset"),
                   ("clone", 'chain cloning depth; one of '
                    '``"deep"``, ``"slim"``'),
                   ("file_changed", "``1`` if the file has changed since last "
                    "save, ``0`` otherwise"),
                   ("power_save", None),
                   ("prelisten", "if non-zero, play notes and instruments "
                    "while entering them"),
                   ("bookmarks", "list of screen bookmarks"),
                   ("wave_synth_overwrite_locks", None)]:
    def field_getter(this):
        return getattr(this.song_data, field)

    def field_setter(this, value):
        setattr(this.song_data, field, value)

    setattr(Song, field, property(fset=field_setter, fget=field_getter,
                                  doc=doc))



from . import bread_spec
import bread
import os
import sys
from . import utils
from .project import Project
from . import blockutils
from .blockutils import BlockReader, BlockWriter, BlockFactory
from . import filepack
import collections
import bitstring
from . import exceptions

from .vendor.six.moves import range

# Start index for data blocks
BLOCKS_START_OFFSET = 0x8000

# The maximum number of files that the .sav can support
NUM_FILES = 0x20

# Constants
EMPTY_BLOCK = 0xff

# By default, SAV file loading doesn't trigger any callback action
def _noop_callback(message, step, total_steps, continuing):
    pass

class ProjectList(object):
    def __init__(self, filename, header_block):
        self.filename = filename
        self.header_block = header_block

        self._projects = {}

    def __eq__(self, other):
        if not isinstance(other, ProjectList):
            return False

        for i in range(NUM_FILES):
            my_proj = self[i]
            other_proj = other[i]

            if my_proj != other_proj:
                return False

        return True

    def __getitem__(self, file_number):
        if file_number not in self._projects:
            with open(self.filename, 'rb') as fp:
                self._projects[file_number] = self._read_project(
                    fp, file_number)

        return self._projects[file_number]

    def __setitem__(self, file_number, proj):
        self._projects[item] = proj

    def keys(self):
        return range(NUM_FILES)

    def _read_project(self, fp, file_number):
        # Scan the block allocation table for blocks for this project
        # Block 0 is reserved, so block numbers are assumed off-by-one
        block_numbers = [
            block_number + 1 for block_number, blk_file in
            enumerate(self.header_block.block_alloc_table)
            if blk_file == file_number
        ]

        project_size_blks = len(block_numbers)

        # If a given file number doesn't have any blocks, it doesn't exist
        if project_size_blks == 0:
            return None

        block_map = {}

        for block_number in block_numbers:
            offset = BLOCKS_START_OFFSET + \
                     (block_number * blockutils.BLOCK_SIZE)

            fp.seek(offset, os.SEEK_SET)

            block_data = bytearray(fp.read(blockutils.BLOCK_SIZE))

            block_map[block_number] = blockutils.Block(
                block_number, block_data)

        reader = BlockReader()
        compressed_data = reader.read(block_map)
        raw_data = filepack.decompress(compressed_data)

        project_name = self.header_block.filenames[file_number]
        project_version = self.header_block.file_versions[file_number]

        project = Project(
            name=self.header_block.filenames[file_number],
            version=self.header_block.file_versions[file_number],
            data=raw_data,
            size_blks=project_size_blks)

        return project

class SAVFile(object):
    # Start offset of SAV file contents
    START_OFFSET = 0x8000

    HEADER_EMPTY_SECTION_1 = (0x8120, 0x813d)

    # Offset of SRAM initialization check
    SRAM_INIT_CHECK_OFFSET = 0x813e
    SRAM_INIT_CHECK_LENGTH = 2

    # Offset where active file number appears
    ACTIVE_FILE_NUMBER_OFFSET = 0x8140

    # Start of block allocation table
    BAT_START_OFFSET = 0x8141

    # End of block allocation table
    BAT_END_OFFSET = 0x81ff

    # Max length in bytes of filename
    FILENAME_LENGTH = 8

    # Length in bytes of file version number
    FILE_VERSION_LENGTH = 1

    # Length in bytes of file number
    FILE_NUMBER_LENGTH = 1

    def __init__(self, filename, callback=_noop_callback):
        """Constructor.

        :param filename: the file to open
        :type name: str
        :param callback: a progress callback function
        :type name: function
        """
        self.filename = filename

        with open(filename, 'rb') as fp:
            self._load(fp, callback)

    def _load(self, fp, callback):
        # read preamble + decompress blocks + "all done"
        total_steps = 3
        current_step = 0

        callback("Reading preamble", current_step, total_steps, True)

        self.preamble = fp.read(self.START_OFFSET)

        header_block_data = fp.read(blockutils.BLOCK_SIZE)

        try:
            self.header_block = bread.parse(
                header_block_data, bread_spec.compressed_sav_file)
        except bitstring.ReadError as e:
            raise exceptions.ImportException(e)

        if self.header_block.sram_init_check != b'jk':
            error_msg = (
                "SRAM init check bits incorrect (should be 'jk', was '%s')" %
                (self.header_block.sram_init_check))

            callback(error_msg, current_step, total_steps, False)
            raise ValueError(error_msg)

        self.active_project_number = self.header_block.active_file

        current_step += 1

        callback("Decompressing", current_step, total_steps, True)

        self.projects = ProjectList(self.filename, self.header_block)

        current_step += 1

        callback("Import complete!", total_steps, total_steps, True)

    def __str__(self):
        output_str = ''

        def add_line(line):
            output_str += line + '\n'

        str_stream = StringIO()

        for i in range(NUM_FILES):
            project = self.projects[i]

            if project is not None:
                add_line(str(project), file=str_stream)

        add_line("Active Project: %s" % \
            (self.projects[self.active_project_number]), file=str_stream)

        return output_str

    def __eq__(self, other):
        return self.projects == other.projects

    @property
    def project_list(self):
        """The list of :py:class:`pylsdj.Project` s that the
        .sav file contains"""
        return [(i, self.projects[i]) for i in sorted(self.projects.keys())]

    def _save(self, fp, callback):
        # Marshal 32 possible projects + write preamble + write data + "all
        # done"
        total_steps = 35
        current_step = 0

        writer = BlockWriter()
        factory = BlockFactory()

        # Block allocation table doesn't include header block because it's
        # always in use, so have to add additional block to account for header
        num_blocks = self.BAT_END_OFFSET - self.BAT_START_OFFSET + 2

        header_block = factory.new_block()

        block_table = []

        for i in range(num_blocks):
            block_table.append(None)

        # First block is the header block, so we should ignore it when creating
        # the block allocation table
        block_table[0] = -1

        for i in range(NUM_FILES):
            project = self.projects[i]

            current_step += 1

            if project is None:
                continue

            callback("Marshaling song '%s'" %
                     (utils.name_without_zeroes(project.name)),
                     current_step - 1, total_steps, True)

            raw_data = project.get_raw_data()
            compressed_data = filepack.compress(raw_data)

            project_block_ids = writer.write(compressed_data, factory)

            for b in project_block_ids:
                block_table[b] = i

        callback("Writing preamble and constructing header block",
                 current_step, total_steps, True)
        current_step += 1
        # Bytes up to START_OFFSET will remain the same
        fp.write(self.preamble)

        # Set header block filenames and versions

        empty_project_name = '\0' * self.FILENAME_LENGTH

        for i in range(NUM_FILES):
            project = self.projects[i]

            if project is None:
                self.header_block.filenames[i] = empty_project_name
                self.header_block.file_versions[i] = 0
            else:
                self.header_block.filenames[i] = project.name
                self.header_block.file_versions[i] = project.version

        self.header_block.active_file = self.active_project_number

        # Ignore the header block when serializing the block allocation table
        for i, b in enumerate(block_table[1:]):
            if b is None:
                file_no = EMPTY_BLOCK
            else:
                file_no = b

            self.header_block.block_alloc_table[i] = file_no

        header_block.data = bread.write(
            self.header_block, bread_spec.compressed_sav_file)

        assert len(header_block.data) == blockutils.BLOCK_SIZE, \
            "Header block isn't the expected length; expected 0x%x, got 0x%x" \
            % (blockutils.BLOCK_SIZE, len(header_block.data))

        block_map = factory.blocks

        empty_block_data = []
        for i in range(blockutils.BLOCK_SIZE):
            empty_block_data.append(0)

        callback("Writing data to file", current_step, total_steps, True)
        current_step += 1
        for i in range(num_blocks):
            if i in block_map:
                data_list = block_map[i].data
            else:
                data_list = empty_block_data

            fp.write(bytearray(data_list))

        callback("Save complete!", total_steps, total_steps, True)

    def save(self, filename, callback=_noop_callback):
        """Save this file.

        :param filename: the file to which to save the .sav file
        :type filename: str
        :param callback: a progress callback function
        :type callback: function
        """
        with open(filename, 'wb') as fp:
            self._save(fp, callback)

if __name__ == "__main__":
    sav = SAVFile(sys.argv[1])
    sav.save(sys.argv[2])



import os
import tempfile
from .vendor.six.moves import range

def printable_decimal_and_hex(num):
    return "{0:d} (0x{0:x})".format(num)

def assert_index_sane(index, upper_bound_exclusive):
    assert type(index) == int, "Indices should be integers; '%s' is not" % (
        index)
    assert 0 <= index < upper_bound_exclusive, (
        "Index %d out of range [%d, %d)" % (index, 0, upper_bound_exclusive))


class ObjectLookupDict(object):

    def __init__(self, id_list, object_list):
        self.id_list = id_list
        self.object_list = object_list

    def __getitem__(self, index):
        assert_index_sane(index, len(self.id_list))

        return self.object_list[self.id_list[index]]

    def __setitem__(self, index, value):
        assert_index_sane(index, len(self.id_list))

        self.id_list[index] = value.index


def name_without_zeroes(name):
    """
    Return a human-readable name without LSDJ's trailing zeroes.

    :param name: the name from which to strip zeroes
    :rtype: the name, without trailing zeroes
    """
    first_zero = name.find(b'\0')

    if first_zero == -1:
        return name
    else:
        return str(name[:first_zero])


class temporary_file:

    def __enter__(self):
        (tmp_handle, tmp_abspath) = tempfile.mkstemp()
        os.close(tmp_handle)
        self.abspath = tmp_abspath
        return self.abspath

    def __exit__(self, t, value, traceback):
        if hasattr(self, 'abspath') and self.abspath is not None:
            os.unlink(self.abspath)



import os
import json

from nose.tools import assert_equal

from .project import load_lsdsng
from .utils import temporary_file

SCRIPT_DIR = os.path.abspath(os.path.dirname(__file__))

def _test_load_store_instrument(source_lsdsng, lsdinst_path, original_index):
    proj = load_lsdsng(source_lsdsng)
    proj.song.instruments.import_from_file(0x2a, lsdinst_path)

    target_instr = proj.song.instruments[0x2a]
    original_instr = proj.song.instruments[original_index]

    assert_equal(original_instr, target_instr)

    with temporary_file() as tmpfile:
        original_instr.export_to_file(tmpfile)

        with open(tmpfile, 'r') as fp:
            saved_inst = json.load(fp)

        with open(lsdinst_path, 'r') as fp:
            original_inst = json.load(fp)

        assert_equal(original_inst, saved_inst)


def test_load_store_wave_instrument():
    _test_load_store_instrument(
        os.path.join(SCRIPT_DIR, 'test_data', 'UNTOLDST.lsdsng'),
        os.path.join(SCRIPT_DIR, 'test_data', 'UNTOLDST_0x00_wave.lsdinst'),
        0x00)

def test_load_store_pulse_instrument():
    _test_load_store_instrument(
        os.path.join(SCRIPT_DIR, 'test_data', 'UNTOLDST.lsdsng'),
        os.path.join(SCRIPT_DIR, 'test_data', 'UNTOLDST_0x03_pulse.lsdinst'),
        0x03)

def test_load_store_kit_instrument():
    _test_load_store_instrument(
        os.path.join(SCRIPT_DIR, 'test_data', 'UNTOLDST.lsdsng'),
        os.path.join(SCRIPT_DIR, 'test_data', 'UNTOLDST_0x16_kit.lsdinst'),
        0x16)

def test_load_store_noise_instrument():
    _test_load_store_instrument(
        os.path.join(SCRIPT_DIR, 'test_data', 'ANNARKTE.lsdsng'),
        os.path.join(SCRIPT_DIR, 'test_data', 'ANNARKTE_0x06_noise.lsdinst'),
        0x06)



import json

from .bread_spec import NUM_SYNTHS, WAVES_PER_SYNTH

class WaveSynthOverwriteLock(object):
    def __init__(self, song, index):
        self._index = NUM_SYNTHS - index - 1
        self._locks = song.song_data.wave_synth_overwrite_locks

    def enable(self):
        self._locks[self._index] = True

    def disable(self):
        self._locks[self._index] = False

    def status(self):
        return self._locks[self._index]

class SynthSoundParams(object):

    def __init__(self, params, overwrite_lock):
        self._params = params
        self._overwrite_lock = overwrite_lock

    def as_native(self):
        return self._params.as_native()

    def import_lsdinst(self, struct):
        self.volume = struct['volume']
        self.filter_cutoff = struct['filter_cutoff']
        self.phase_amount = struct['phase_amount']
        self.vertical_shift = struct['vertical_shift']

    @property
    def volume(self):
        """the wave's volume"""
        return self._params.volume

    @volume.setter
    def volume(self, value):
        self._params.volume = value
        self._overwrite_lock.disable()

    @property
    def filter_cutoff(self):
        """the filter's cutoff frequency"""
        return self._params.filter_cutoff

    @filter_cutoff.setter
    def filter_cutoff(self, value):
        self._params.filter_cutoff = value
        self._overwrite_lock.disable()

    @property
    def phase_amount(self):
        """the amount of phase shift, ``0`` = no phase,
        ``0x1f`` = maximum phase"""
        return self._params.phase_amount

    @phase_amount.setter
    def phase_amount(self, value):
        self._params.phase_amount = value
        self._overwrite_lock.disable()

    @property
    def vertical_shift(self):
        """the amount to shift the waveform vertically"""
        return self._params.vertical_shift

    @vertical_shift.setter
    def vertical_shift(self, value):
        self._params.vertical_shift = value
        self._overwrite_lock.disable()

class WaveFrames(object):
    def __init__(self, song, synth_index, wave_index, overwrite_lock):
        self._frames = song.song_data.wave_frames[synth_index][wave_index]
        self._overwrite_lock = overwrite_lock

    def __getitem__(self, index):
        return self._frames[index]

    def __setitem__(self, index, value):
        self._frames[index] = value
        self._overwrite_lock.enable()

class Waves(object):
    def __init__(self, song, synth_index, overwrite_lock):
        self._waves = [WaveFrames(song, synth_index, wave_index, overwrite_lock)
                       for wave_index in range(WAVES_PER_SYNTH)]

    def __getitem__(self, index):
        return self._waves[index]

class Synth(object):

    def __init__(self, song, index):
        self._song = song
        self._index = index

        self._overwrite_lock = WaveSynthOverwriteLock(song, index)

        self._params = self._song.song_data.softsynth_params[index]

        self._start = SynthSoundParams(
            self._song.song_data.softsynth_params[index].start,
            self._overwrite_lock)
        self._end = SynthSoundParams(
            self._song.song_data.softsynth_params[index].end,
            self._overwrite_lock)

        self._waves = Waves(song, index, self._overwrite_lock)

    @property
    def song(self):
        """the synth's parent Song"""
        return self._song

    @property
    def index(self):
        """the synth's index within its parent song's synth table"""
        return self._index

    @property
    def start(self):
        """parameters for the start of the sound, represented as a
        SynthSoundParams object"""
        return self._start

    @property
    def end(self):
        """parameters for the end of the sound, represented as a
        SynthSoundParams object"""
        return self._end

    @property
    def waveform(self):
        '''the synth\'s waveform type; one of ``"sawtooth"``,
        ``"square"``, ``"sine"``'''
        return self._params.waveform

    @waveform.setter
    def waveform(self, value):
        self._params.waveform = value
        self._overwrite_lock.disable()

    @property
    def filter_type(self):
        '''the type of filter applied to the waveform; one of
        ``"lowpass"``, ``"highpass"``, ``"bandpass"``, ``"allpass"``'''
        return self._params.filter_type

    @filter_type.setter
    def filter_type(self, value):
        self._params.filter_type = value
        self._overwrite_lock.disable()

    @property
    def filter_resonance(self):
        """boosts the signal around the cutoff
        frequency, to change how bright or dull the wave sounds"""
        return self._params.filter_resonance

    @filter_resonance.setter
    def filter_resonance(self, value):
        self._params.filter_resonance = value
        self._overwrite_lock.disable()

    @property
    def distortion(self):
        '''use ``"clip"`` or ``"wrap"`` distortion'''
        return self._params.distortion

    @distortion.setter
    def distortion(self, value):
        self._params.distortion = value
        self._overwrite_lock.disable()

    @property
    def phase_type(self):
        '''compresses the waveform horizontally; one of
        ``"normal"``, ``"resync"``, ``"resync2"``'''
        return self._params.phase_type

    @phase_type.setter
    def phase_type(self, value):
        '''compresses the waveform horizontally; one of
        ``"normal"``, ``"resync"``, ``"resync2"``'''
        self._params.phase_type = value
        self._overwrite_lock.disable()

    @property
    def waves(self):
        """a list of the synth's waveforms, each of which is a list of bytes"""
        return self._waves

    @property
    def wave_synth_overwrite_lock(self):
        """if True, the synth's waveforms override its synth parameters;
        if False, its synth parameters override its waveforms"""
        return self._overwrite_lock.status()

    def export(self):
        export_struct = {}

        export_struct["params"] = json.loads(self._params.as_json())
        export_struct["waves"] = []

        for wave in self.waves:
            export_struct["waves"].append(list(wave))

        return export_struct

    def import_lsdinst(self, synth_data):
        import_keys = ['start', 'end', 'waveform', 'filter_type',
                       'filter_resonance', 'distortion', 'phase_type']

        for key in import_keys:
            value = synth_data['params'][key]

            if key in ('start', 'end'):
                getattr(self, key).import_lsdinst(value)
            else:
                setattr(self, key, value)

        for i, wave in enumerate(synth_data['waves']):
            for j, frame in enumerate(wave):
                self.waves[i][j] = frame



import os
import sys
import json
from nose.tools import raises, assert_equal, assert_list_equal
from .vendor.six.moves import range

SCRIPT_DIR = os.path.abspath(os.path.dirname(__file__))

sys.path.append(os.path.join(SCRIPT_DIR, os.path.pardir))


def assert_bytearray_equal(x, y):
    assert_list_equal(list(map(hex, x)), list(map(hex, y)))

from . import filepack as filepack
from . import blockutils as bl
from . import bread_spec as bread_spec
import bread as b


def test_basic_compress_decompress():
    data = [i % 10 for i in range(5000)]

    compressed = filepack.compress(data)

    decompressed = filepack.decompress(compressed)

    assert_list_equal(data, decompressed)


def test_rle_compress():
    data = [0xde for i in range(350)]
    data.extend([0xfe for i in range(220)])
    data.append(42)
    data.append(17)

    compressed = filepack.compress(data)

    reference = [filepack.RLE_BYTE, 0xde, 255,
                 filepack.RLE_BYTE, 0xde, 95,
                 filepack.RLE_BYTE, 0xfe, 220,
                 42, 17]

    assert_list_equal(compressed, reference)

    decompressed = filepack.decompress(compressed)

    assert_list_equal(decompressed, data)


def test_short_rle_compress():
    data = [0xde, 0xde, 42, 17, 12]

    compressed = filepack.compress(data)

    assert_list_equal(compressed, data)


def test_rle_special_byte():
    data = [filepack.RLE_BYTE, filepack.RLE_BYTE,
            filepack.SPECIAL_BYTE, filepack.SPECIAL_BYTE,
            filepack.RLE_BYTE, filepack.SPECIAL_BYTE]

    reference = [filepack.RLE_BYTE,
                 filepack.RLE_BYTE,
                 filepack.RLE_BYTE,
                 filepack.RLE_BYTE,
                 filepack.SPECIAL_BYTE,
                 filepack.SPECIAL_BYTE,
                 filepack.SPECIAL_BYTE,
                 filepack.SPECIAL_BYTE,
                 filepack.RLE_BYTE,
                 filepack.RLE_BYTE,
                 filepack.SPECIAL_BYTE,
                 filepack.SPECIAL_BYTE]

    compressed = filepack.compress(data)

    assert_list_equal(compressed, reference)

    decompressed = filepack.decompress(compressed)

    assert_list_equal(decompressed, data)


def test_default_instr_compress():
    data = []

    for i in range(42):
        data.extend(filepack.DEFAULT_INSTRUMENT)

    reference = []

    reference.extend([0, filepack.SPECIAL_BYTE,
                      filepack.DEFAULT_INSTR_BYTE, 41])
    reference.extend(filepack.DEFAULT_INSTRUMENT_FILEPACK)

    reference = bytearray(reference)

    decompressed_reference = filepack.decompress(reference)
    assert_list_equal(decompressed_reference, data + [0])

    data = bytearray(data)

    compressed = filepack.compress(data)

    assert_bytearray_equal(bytearray(compressed + [0]), reference)

    decompressed = filepack.decompress(compressed)

    assert_bytearray_equal(data, bytearray(decompressed))


def test_instrument_sizes():
    instr_bytes = list(filepack.DEFAULT_INSTRUMENT)

    parsed = b.parse(instr_bytes, bread_spec.instrument)
    assert_equal(len(parsed), 16 * 8)

    raw_data = b.write(parsed, bread_spec.instrument)
    assert_bytearray_equal(raw_data, bytearray(filepack.DEFAULT_INSTRUMENT))

    instr_bytes[0] = 1

    parsed = b.parse(instr_bytes, bread_spec.instrument)
    assert_equal(len(parsed), 16 * 8)

    instr_bytes[0] = 2

    parsed = b.parse(instr_bytes, bread_spec.instrument)
    assert_equal(len(parsed), 16 * 8)

    instr_bytes[0] = 3

    parsed = b.parse(instr_bytes, bread_spec.instrument)
    assert_equal(len(parsed), 16 * 8)


def test_default_wave_compress():
    data = []

    for i in range(33):
        data.extend(filepack.DEFAULT_WAVE)

    compressed = filepack.compress(data)

    reference = [filepack.SPECIAL_BYTE, filepack.DEFAULT_WAVE_BYTE, 33]

    decompressed_reference = filepack.decompress(reference)
    assert_list_equal(decompressed_reference, data)

    decompressed = filepack.decompress(compressed)

    assert_list_equal(data, decompressed)
    assert_list_equal(compressed, reference)


def test_large_rle_compress():
    data = []

    for i in range(275):
        data.append(42)

    compressed = filepack.compress(data)

    reference = [filepack.RLE_BYTE, 42, 255, filepack.RLE_BYTE, 42, 20]

    assert_list_equal(compressed, reference)

    decompressed = filepack.decompress(compressed)

    assert_list_equal(data, decompressed)


@raises(AssertionError)
def test_bad_rle_split():
    data = [filepack.RLE_BYTE]

    factory = bl.BlockFactory()

    filepack.split(data, bl.BLOCK_SIZE, factory)


@raises(AssertionError)
def test_bad_special_byte_split():
    data = [filepack.SPECIAL_BYTE]

    factory = bl.BlockFactory()

    filepack.split(data, bl.BLOCK_SIZE, factory)


@raises(AssertionError)
def test_block_jump_during_split_asserts():
    data = [filepack.SPECIAL_BYTE, 47]

    factory = bl.BlockFactory()

    filepack.split(data, bl.BLOCK_SIZE, factory)


def test_special_byte_on_block_boundary():
    data = [42, 17, filepack.SPECIAL_BYTE, filepack.SPECIAL_BYTE,
            100, 36]

    factory = bl.BlockFactory()

    filepack.split(data, 5, factory)

    block_0_expected = [42, 17, filepack.SPECIAL_BYTE, 1, 0]
    block_1_expected = [filepack.SPECIAL_BYTE, filepack.SPECIAL_BYTE, 100,
                        filepack.SPECIAL_BYTE, 2]
    block_2_expected = [36, filepack.SPECIAL_BYTE, filepack.EOF_BYTE, 0, 0]

    assert_equal(len(factory.blocks), 3)
    assert_list_equal(factory.blocks[0].data, block_0_expected)
    assert_list_equal(factory.blocks[1].data, block_1_expected)
    assert_list_equal(factory.blocks[2].data, block_2_expected)


def test_rle_byte_on_block_boundary():
    data = [42, 17, filepack.RLE_BYTE, filepack.RLE_BYTE,
            100, 36]

    factory = bl.BlockFactory()

    filepack.split(data, 5, factory)

    block_0_expected = [42, 17, filepack.SPECIAL_BYTE, 1, 0]
    block_1_expected = [filepack.RLE_BYTE, filepack.RLE_BYTE, 100,
                        filepack.SPECIAL_BYTE, 2]
    block_2_expected = [36, filepack.SPECIAL_BYTE, filepack.EOF_BYTE, 0, 0]

    assert_equal(len(factory.blocks), 3)
    assert_list_equal(factory.blocks[0].data, block_0_expected)
    assert_list_equal(factory.blocks[1].data, block_1_expected)
    assert_list_equal(factory.blocks[2].data, block_2_expected)


def test_full_rle_on_block_boundary():
    data = [42, filepack.RLE_BYTE, 55, 4, 22, 3]

    factory = bl.BlockFactory()

    filepack.split(data, 5, factory)

    block_0_expected = [42, filepack.SPECIAL_BYTE, 1, 0, 0]
    block_1_expected = [filepack.RLE_BYTE, 55, 4, filepack.SPECIAL_BYTE, 2]
    block_2_expected = [22, 3, filepack.SPECIAL_BYTE, filepack.EOF_BYTE, 0]

    assert_equal(len(factory.blocks), 3)
    assert_list_equal(factory.blocks[0].data, block_0_expected)
    assert_list_equal(factory.blocks[1].data, block_1_expected)
    assert_list_equal(factory.blocks[2].data, block_2_expected)


def test_default_on_block_boundary():
    data = [42, filepack.SPECIAL_BYTE, filepack.DEFAULT_INSTR_BYTE, 3, 2, 5]

    factory = bl.BlockFactory()

    filepack.split(data, 5, factory)

    block_0_expected = [42, filepack.SPECIAL_BYTE, 1, 0, 0]
    block_1_expected = [filepack.SPECIAL_BYTE, filepack.DEFAULT_INSTR_BYTE, 3,
                        filepack.SPECIAL_BYTE, 2]
    block_2_expected = [2, 5, filepack.SPECIAL_BYTE, filepack.EOF_BYTE, 0]

    assert_equal(len(factory.blocks), 3)
    assert_list_equal(factory.blocks[0].data, block_0_expected)
    assert_list_equal(factory.blocks[1].data, block_1_expected)
    assert_list_equal(factory.blocks[2].data, block_2_expected)


def test_merge_with_rle_byte():
    factory = bl.BlockFactory()

    block1 = factory.new_block()
    block1.data = [filepack.RLE_BYTE, filepack.RLE_BYTE, 2, 1, 3,
                   filepack.SPECIAL_BYTE, 1, 0, 0]
    block2 = factory.new_block()
    block2.data = [4, 3, 6, filepack.SPECIAL_BYTE, filepack.EOF_BYTE]

    data = filepack.merge(factory.blocks)
    expected_data = [filepack.RLE_BYTE, filepack.RLE_BYTE, 2, 1, 3, 4, 3, 6]

    assert_list_equal(data, expected_data)


def test_merge_with_full_rle():
    factory = bl.BlockFactory()
    block1 = factory.new_block()
    block1.data = [filepack.RLE_BYTE, 42, 17, 1, 1, 4,
                   filepack.SPECIAL_BYTE, 1, 0, 0]
    block2 = factory.new_block()
    block2.data = [4, 4, 42, filepack.SPECIAL_BYTE, filepack.EOF_BYTE]

    data = filepack.merge(factory.blocks)

    assert_list_equal(data, [filepack.RLE_BYTE, 42, 17, 1, 1, 4, 4, 4, 42])


def test_merge_with_special_byte():
    factory = bl.BlockFactory()

    block1 = factory.new_block()
    block1.data = [filepack.SPECIAL_BYTE, filepack.SPECIAL_BYTE, 2, 1, 3,
                   filepack.SPECIAL_BYTE, 1, 0, 0]
    block2 = factory.new_block()
    block2.data = [4, 3, 6, filepack.SPECIAL_BYTE, filepack.EOF_BYTE]

    data = filepack.merge(factory.blocks)
    expected_data = [filepack.SPECIAL_BYTE, filepack.SPECIAL_BYTE, 2, 1, 3, 4,
                     3, 6]

    assert_list_equal(data, expected_data)


def test_merge_with_special_command():
    factory = bl.BlockFactory()

    block1 = factory.new_block()
    block1.data = [filepack.SPECIAL_BYTE, filepack.DEFAULT_INSTR_BYTE, 4, 6,
                   1, 93, filepack.SPECIAL_BYTE, 1, 0, 0]
    block2 = factory.new_block()
    block2.data = [3, 3, 33, filepack.SPECIAL_BYTE, filepack.EOF_BYTE]

    data = filepack.merge(factory.blocks)
    expected_data = [filepack.SPECIAL_BYTE, filepack.DEFAULT_INSTR_BYTE, 4, 6,
                     1, 93, 3, 3, 33]

    assert_list_equal(data, expected_data)


@raises(AssertionError)
def test_decompress_bogus_special_byte_asserts():
    data = [filepack.SPECIAL_BYTE, filepack.EOF_BYTE]

    filepack.decompress(data)


def test_weird_rle_compress():
    data = [0x1b, 0xc0, 0x00, 0x0f, 0x1d, 0xc0, 0x00, 0x0f, 0x1e, 0xc0,
            0x00, 0x0f, 0x20, 0xc0, 0x00, 0x1f, 0x22, 0x00, 0x00, 0x00,
            0x1e, 0x00, 0x1b, 0x00, 0x00, 0x00, 0x1e, 0x00, 0x00, 0x00,
            0x1d, 0x00, 0x00, 0x00, 0x19, 0x00, 0x00, 0x00, 0x20, 0x00,
            0x00, 0x00, 0x20, 0x00, 0x1e, 0x00, 0x1d, 0x00, 0x1b, 0x00,
            0x00, 0x00, 0x17, 0x00, 0x19, 0x00, 0x00, 0x00, 0x1b, 0x00,
            0x00, 0x00, 0x1e, 0x00, 0x1d, 0x00, 0x00, 0x00, 0x19, 0x00,
            0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x1d, 0x00, 0x00, 0x00,
            0x1e, 0x00, 0x00, 0x00, 0x1b, 0x00, 0x1e, 0x00, 0x00, 0x00,
            0x22, 0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x1d, 0x00,
            0x00, 0x00, 0x25, 0x00, 0x00, 0x00, 0x20, 0xc0, 0x00, 0x05,
            0x23, 0x00, 0x00, 0x00, 0x25, 0x00, 0x23, 0x00, 0x00, 0x00,
            0x1e, 0xc0, 0x00, 0x05, 0x20, 0x00, 0x00, 0x00, 0x1d, 0x00,
            0x19, 0x00, 0x00, 0x00, 0x1d, 0xc0, 0x00, 0x05, 0x20, 0x00,
            0x00, 0x00, 0x1d, 0x00, 0x19, 0x00, 0x00, 0x00, 0x1d, 0x00,
            0x00, 0x00, 0x1b, 0x00, 0x1b, 0x00, 0x00, 0x00, 0x16, 0x00,
            0x00, 0x00, 0x1b, 0x00, 0x00, 0x00, 0x1e, 0x00, 0x00, 0x00,
            0x1d, 0x00, 0x00, 0x00, 0x19, 0x00, 0x20, 0x00, 0x00, 0x00,
            0x1d, 0xc0, 0x00, 0x05, 0x1b, 0x00, 0x00, 0x00, 0x03, 0x00,
            0x03, 0x00, 0x00, 0x00, 0x03, 0x00, 0x03, 0x00, 0x1b, 0x19,
            0x03, 0xc0, 0x00, 0x0d, 0x03, 0x01, 0x0f, 0xc0, 0x00, 0x0f,
            0x19, 0xc0, 0x00, 0x1f, 0x1e, 0x00, 0x00, 0x00, 0x23, 0x00,
            0x1e, 0x00, 0x00, 0x00, 0x1b, 0xc0, 0x00, 0x05, 0x1d, 0x00,
            0x00, 0x00, 0x20, 0x00, 0x1d, 0x00, 0x00, 0x00, 0x19, 0xc0,
            0x00, 0x05, 0x1b, 0xc0, 0x00, 0x8f, 0x31, 0x00, 0x1b, 0x00,
            0x31, 0x00, 0x1b, 0x00, 0x31, 0x00, 0x1b, 0x00, 0x31, 0x00,
            0x1b, 0x00, 0x31, 0x00, 0x19, 0x00, 0x31, 0x00, 0x19, 0x00,
            0x31, 0x00, 0x19, 0x00, 0x31, 0x00, 0x19, 0x00, 0x31, 0x00,
            0x17, 0x00, 0x31, 0x00, 0x17, 0x00, 0x31, 0x00, 0x17, 0x00,
            0x31, 0x00, 0x17, 0x00, 0x31, 0xc0, 0x00, 0x09, 0x31, 0xc0,
            0x00, 0x05, 0x31, 0x00, 0x31, 0xc0, 0x00, 0x07, 0x31, 0xc0,
            0x00, 0x05, 0x31, 0x00, 0x0f, 0x00, 0x31, 0x00, 0x0f, 0x00,
            0x31, 0x00, 0x0f, 0x00, 0x31, 0x00, 0x0f, 0x00, 0x31, 0xc0,
            0x00, 0x9f, 0x01, 0x00, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00,
            0x01, 0x01, 0x01, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01, 0x00,
            0x01, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01, 0x01, 0x01, 0x00,
            0x01, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01, 0x00,
            0x01, 0x00, 0xc0, 0x01, 0x08, 0x00, 0x00, 0xc0, 0x01, 0x0e,
            0xc0, 0x00, 0x10, 0x01, 0xc0, 0x00, 0xff, 0xc0, 0x00, 0xff,
            0xc0, 0x00, 0xff, 0xc0, 0x00, 0xff, 0xc0, 0x00, 0xff, 0xc0,
            0x00, 0xff, 0xc0, 0x00, 0xff, 0xc0, 0x00, 0xff, 0xc0, 0x00,
            0xff, 0xc0, 0x00, 0xff, 0xc0, 0x00, 0xff, 0xc0, 0x00, 0xff,
            0xc0, 0x00, 0xab, 0xc0, 0xff, 0x40, 0xc0, 0x00, 0x60, 0x06,
            0x06, 0xc0, 0x00, 0x0e, 0x06, 0x06, 0xc0, 0x00, 0x0e, 0x06,
            0x06, 0xc0, 0x00, 0x0e, 0x06, 0x06, 0xc0, 0x00, 0x0e, 0x06,
            0x06, 0xc0, 0x00, 0x0e, 0x06, 0x06, 0xc0, 0x00, 0x0e, 0x06,
            0x06, 0xc0, 0x00, 0x0e, 0x06, 0x06, 0xc0, 0x00, 0x0e, 0x06,
            0x06, 0xc0, 0x00, 0x0e, 0x06, 0x06, 0xc0, 0x00, 0x0e, 0x06,
            0x06, 0xc0, 0x00, 0x0e, 0x06, 0x06, 0xc0, 0x00, 0x0e, 0x06,
            0x06, 0xc0, 0x00, 0x0e, 0x06, 0x06, 0xc0, 0x00, 0x0e, 0x06,
            0x06, 0xc0, 0x00, 0x0e, 0x06, 0x06, 0xc0, 0x00, 0x0e, 0x06,
            0x06, 0xc0, 0x00, 0x0e, 0x06, 0x06, 0xc0, 0x00,
            0x0e, 0x06, 0x06, 0xc0, 0x00, 0x0e, 0x06, 0x06, 0xc0, 0x00,
            0x0e, 0x06, 0x06, 0xc0, 0x00, 0x0e, 0x06, 0x06, 0xc0, 0x00,
            0x0e, 0x06, 0x06, 0xc0, 0x00, 0x0e, 0x06, 0x06, 0xc0, 0x00,
            0x0e, 0x06, 0x06, 0xc0, 0x00, 0x0e, 0x06, 0x06, 0xc0, 0x00,
            0x0e, 0x06, 0x06, 0xc0, 0x00, 0x0e, 0x06, 0x06, 0xc0, 0x00,
            0x0e, 0x06, 0x06, 0xc0, 0x00, 0x0e, 0x06, 0x06, 0xc0, 0x00,
            0x0e, 0x06, 0x06, 0xc0, 0x00, 0x0e, 0x06, 0x06, 0xc0, 0x00,
            0x0e, 0x7f, 0x7f, 0x20, 0x31, 0x7f, 0x7f, 0x20, 0x30, 0x7f,
            0x7f, 0x20, 0x30, 0x7f, 0x7f, 0x20, 0x30, 0x7f, 0x10, 0x21,
            0x30, 0x7f, 0x10, 0x21, 0x30, 0x7f, 0x7f, 0x20, 0x30, 0x7f,
            0x7f, 0x20, 0x30, 0x7f, 0x10, 0x21, 0x30, 0x7f, 0x10, 0x21,
            0x30, 0x7f, 0x7f, 0x20, 0x30, 0x7f, 0x7f, 0x20, 0x30, 0x7f,
            0x11, 0x22, 0x31, 0x7f, 0x12, 0x22, 0x30, 0x7f, 0x13, 0x22,
            0x30, 0x7f, 0x12, 0x22, 0x30, 0x7f, 0x7f, 0x20, 0x30, 0x7f,
            0x7f, 0x20, 0x30, 0x7f, 0x7f, 0x20, 0x30, 0x7f, 0x7f, 0x20,
            0x30, 0x7e, 0x7e, 0x23, 0x32, 0xc0, 0xff, 0xff, 0xc0, 0xff,
            0xff, 0xc0, 0xff, 0xff, 0xc0, 0xff, 0xaf, 0xc0, 0x00, 0xff,
            0xc0, 0x00, 0xff, 0xc0, 0x00, 0xff, 0xc0, 0x00, 0xff, 0xc0,
            0x00, 0xff, 0xc0, 0x00, 0xff, 0xc0, 0x00, 0xff, 0xc0, 0x00,
            0x47, 0x43, 0x20, 0x33, 0x20, 0x43, 0x23, 0x33, 0x20, 0x44,
            0x20, 0x33, 0x20, 0x44, 0x23, 0x33, 0x20, 0x45, 0x20, 0x33,
            0x20, 0x46, 0x20, 0x33, 0x20, 0x46, 0x23, 0x33, 0x20, 0x47,
            0x20, 0x33, 0x20, 0x47, 0x23, 0x33, 0x20, 0x41, 0x20, 0x33,
            0x20, 0x41, 0x23, 0x33, 0x20, 0x42, 0x20, 0x33, 0x20, 0x43,
            0x20, 0x34, 0x20, 0x43, 0x23, 0x34, 0x20, 0x44, 0x20, 0x34,
            0x20, 0x44, 0x23, 0x34, 0x20, 0x45, 0x20, 0x34, 0x20, 0x46,
            0x20, 0x34, 0x20, 0x46, 0x23, 0x34, 0x20, 0x47, 0x20, 0x34,
            0x20, 0x47, 0x23, 0x34, 0x20, 0x41, 0x20, 0x34, 0x20, 0x41,
            0x23, 0x34, 0x20, 0x42, 0x20, 0x34, 0x20, 0x43, 0x20, 0x35,
            0x20, 0x43, 0x23, 0x35, 0x20, 0x44, 0x20, 0x35, 0x20, 0x44,
            0x23, 0x35, 0x20, 0x45, 0x20, 0x35, 0x20, 0x46, 0x20, 0x35,
            0x20, 0x46, 0x23, 0x35, 0x20, 0x47, 0x20, 0x35, 0x20, 0x47,
            0x23, 0x35, 0x20, 0x41, 0x20, 0x35, 0x20, 0x41, 0x23, 0x35,
            0x20, 0x42, 0x20, 0x35, 0x20, 0x43, 0x20, 0x36, 0x20, 0x43,
            0x23, 0x36, 0x20, 0x44, 0x20, 0x36, 0x20, 0x44, 0x23, 0x36,
            0x20, 0x45, 0x20, 0x36, 0x20, 0x46, 0x20, 0x36, 0x20, 0x72,
            0x62, 0xc0, 0x00, 0xff, 0xc0, 0x00, 0xa7, 0xc0, 0x01, 0x09,
            0xc0, 0x00, 0x17, 0x01, 0x01, 0x01, 0xc0, 0x00, 0x0d, 0x01,
            0x01, 0x01, 0xc0, 0x00, 0x0d, 0x01, 0x01, 0xc0, 0x00, 0x0e,
            0xc0, 0x01, 0x05, 0xc0, 0x00, 0x0c, 0x01, 0x02, 0x02, 0xc0,
            0xff, 0x0c, 0x05, 0x06, 0x07, 0x08, 0xc0, 0xff, 0x0c, 0x09,
            0x0a, 0x0b, 0x0c, 0xc0, 0xff, 0x0c, 0x02, 0x03, 0x00, 0x00,
            0xc0, 0xff, 0x0c, 0x09, 0x0a, 0x0b, 0x0d, 0xc0, 0xff, 0x0c,
            0x0e, 0x0f, 0x15, 0x16, 0xc0, 0xff, 0x0c, 0x17, 0xc0, 0xfe,
            0x0f, 0xc0, 0xff, 0x90, 0xc0, 0x11, 0x04, 0xc0, 0xff, 0x0c,
            0x12, 0x12, 0x13, 0x13, 0xc0, 0xff, 0x0c, 0xc0, 0x13, 0x04,
            0xc0, 0xff, 0x0c, 0x12, 0x12, 0x13, 0x13, 0xc0, 0xff, 0xcc,
            0x20, 0x21, 0x22, 0x21, 0xc0, 0xff, 0x0c, 0x23, 0x24, 0x23,
            0x24, 0xc0, 0xff, 0x0c, 0xc0, 0x25, 0x04, 0xc0, 0xff, 0x0c,
            0x26, 0xc0, 0xfe, 0x0f, 0xc0, 0xff, 0xc0, 0x30, 0x31, 0x31,
            0x32, 0x00, 0xc0, 0xff, 0x0c, 0xfe, 0xfe, 0xfe,
            0x33, 0xc0, 0xff, 0x0c, 0x35, 0xc0, 0xfe, 0x0f, 0xc0, 0xff,
            0xff, 0xc0, 0xff, 0xff, 0xc0, 0xff, 0xff, 0xc0, 0xff, 0xff,
            0xc0, 0xff, 0xb4, 0xc0, 0xfe, 0x0b, 0xee, 0xee, 0xc0, 0xfe,
            0x07, 0xc0, 0xff, 0x0c, 0x00, 0x00, 0x00, 0x02, 0xc0, 0x00,
            0x2c, 0x04, 0x00, 0x08, 0x0a, 0xc0, 0x00, 0xec, 0xfe, 0xfe,
            0xc0, 0x00, 0x0e, 0x0c, 0x0c, 0xc0, 0x00, 0xff, 0xc0, 0x00,
            0xff, 0xc0, 0x00, 0xff, 0xc0, 0x00, 0xff, 0xc0, 0x00, 0xff,
            0xc0, 0x00, 0xff, 0xc0, 0x00, 0xd5, 0x88, 0x00, 0x3f, 0xff,
            0x00, 0x22, 0x83, 0x00, 0x00, 0xd0, 0x00, 0x00, 0x00, 0xf3]

    decompressed = filepack.decompress(data)

    recompressed = filepack.compress(decompressed)
    assert_list_equal(data, recompressed)


def test_sample_song():
    sample_song_compressed = os.path.join(
        SCRIPT_DIR, "test_data", "sample_song_compressed.json")

    with open(sample_song_compressed, "r") as fp:
        compressed = json.load(fp)

    decompressed = filepack.decompress(compressed)

    recompressed = filepack.compress(decompressed)

    assert_equal(len(decompressed), 0x8000)

    assert_equal(len(recompressed), len(compressed))
    assert_list_equal(recompressed, compressed)



__title__ = 'pylsdj'
__version__ = '2.3.2'
__author__ = 'Alex Rasmussen'
__license__ = 'MIT'
__copyright__ = 'Copyright 2015 Alex Rasmussen'

from . import bread_spec
from . import chain
from . import clock
from . import consts
from . import filepack
from .instrument import Instrument
from .wave_instrument import WaveInstrument
from .pulse_instrument import PulseInstrument
from .kit_instrument import KitInstrument
from .noise_instrument import NoiseInstrument
from .phrase import Phrase
from .project import load_lsdsng, load_srm, Project
from .savfile import SAVFile
from .song import Song, Sequence, Instruments
from .speech_instrument import Word, SpeechInstrument
from .synth import Synth, SynthSoundParams
from .table import Table, TableFX
from .vibrato import Vibrato
from . import exceptions
from . import utils



import itertools

from .vendor.six.moves import range

# Byte used to denote run-length encoding
RLE_BYTE = 0xc0

# Byte used to denote special action
SPECIAL_BYTE = 0xe0

# Byte used to denote end of file (appears after special byte)
EOF_BYTE = 0xff

# Byte used to denote default instrument
DEFAULT_INSTR_BYTE = 0xf1

# Byte used to denote default wave
DEFAULT_WAVE_BYTE = 0xf0

DEFAULT_WAVE = bytearray(
    [0x8e, 0xcd, 0xcc, 0xbb, 0xaa, 0xa9, 0x99, 0x88, 0x87, 0x76,
     0x66, 0x55, 0x54, 0x43, 0x32, 0x31])

DEFAULT_INSTRUMENT_FILEPACK = bytearray([
    0xa8, 0, 0, 0xff, 0, 0, 3, 0, 0, 0xd0, 0, 0, 0, 0xf3, 0, 0])

DEFAULT_INSTRUMENT = bytearray([
    0, 0xa8, 0, 0, 0xff, 0, 0, 3, 0, 0, 0xd0, 0, 0, 0, 0xf3, 0])

# DEFAULT_INSTRUMENT = [
#     0xa8, 0, 0, 0xff, 0, 0, 3, 0, 0, 0xd0, 0, 0, 0, 0xf3, 0, 0]

RESERVED_BYTES = [SPECIAL_BYTE, RLE_BYTE]

SPECIAL_DEFAULTS = [DEFAULT_INSTR_BYTE, DEFAULT_WAVE_BYTE]

STATE_BYTES = 0
STATE_RLE_BYTE = 1
STATE_RLE_COUNT = 2
STATE_SPECIAL_BYTE = 3
STATE_DEFAULT_INSTR = 4
STATE_DEFAULT_WAVE = 5
STATE_DONE = 6


def split(compressed_data, segment_size, block_factory):
    """Splits compressed data into blocks.

    :param compressed_data: the compressed data to split
    :param segment_size: the size of a block in bytes
    :param block_factory: a BlockFactory used to construct the blocks

    :rtype: a list of block IDs of blocks that the block factory created while
    splitting
    """

    # Split compressed data into blocks
    segments = []

    current_segment_start = 0
    index = 0
    data_size = len(compressed_data)

    while index < data_size:
        current_byte = compressed_data[index]

        if index < data_size - 1:
            next_byte = compressed_data[index + 1]
        else:
            next_byte = None

        jump_size = 1

        if current_byte == RLE_BYTE:
            assert next_byte is not None, "Expected a command to follow " \
                "RLE byte"
            if next_byte == RLE_BYTE:
                jump_size = 2
            else:
                jump_size = 3

        elif current_byte == SPECIAL_BYTE:
            assert next_byte is not None, "Expected a command to follow " \
                "special byte"

            if next_byte == SPECIAL_BYTE:
                jump_size = 2
            elif next_byte == DEFAULT_INSTR_BYTE or \
                    next_byte == DEFAULT_WAVE_BYTE:
                jump_size = 3
            else:
                assert False, "Encountered unexpected EOF or block " \
                    "switch while segmenting"

        # Need two bytes for the jump or EOF
        if index - current_segment_start + jump_size > segment_size - 2:
            segments.append(compressed_data[
                current_segment_start:index])

            current_segment_start = index
        else:
            index += jump_size

    # Append the last segment, if any
    if current_segment_start != index:
        segments.append(compressed_data[
            current_segment_start:current_segment_start + index])

    # Make sure that no data was lost while segmenting
    total_segment_length = sum(map(len, segments))
    assert total_segment_length == len(compressed_data), "Lost %d bytes of " \
        "data while segmenting" % (len(compressed_data) - total_segment_length)

    block_ids = []

    for segment in segments:
        block = block_factory.new_block()
        block_ids.append(block.id)

    for (i, segment) in enumerate(segments):
        block = block_factory.blocks[block_ids[i]]

        assert len(block.data) == 0, "Encountered a block with "
        "pre-existing data while writing"

        if i == len(segments) - 1:
            # Write EOF to the end of the segment
            add_eof(segment)
        else:
            # Write a pointer to the next segment
            add_block_switch(segment, block_ids[i + 1])

        # Pad segment with zeroes until it's large enough
        pad(segment, segment_size)

        block.data = segment

    return block_ids


def renumber_block_keys(blocks):
    """Renumber a block map's indices so that tehy match the blocks' block
    switch statements.

    :param blocks a block map to renumber
    :rtype: a renumbered copy of the block map
    """

    # There is an implicit block switch to the 0th block at the start of the
    # file
    byte_switch_keys = [0]
    block_keys = list(blocks.keys())

    # Scan the blocks, recording every block switch statement
    for block in list(blocks.values()):
        i = 0
        while i < len(block.data) - 1:
            current_byte = block.data[i]
            next_byte = block.data[i + 1]

            if current_byte == RLE_BYTE:
                if next_byte == RLE_BYTE:
                    i += 2
                else:
                    i += 3
            elif current_byte == SPECIAL_BYTE:
                if next_byte in SPECIAL_DEFAULTS:
                    i += 3
                elif next_byte == SPECIAL_BYTE:
                    i += 2
                else:
                    if next_byte != EOF_BYTE:
                        byte_switch_keys.append(next_byte)

                    break

            else:
                i += 1

    byte_switch_keys.sort()
    block_keys.sort()

    assert len(byte_switch_keys) == len(block_keys), (
        "Number of blocks that are target of block switches (%d) "
        % (len(byte_switch_keys)) +
        "does not equal number of blocks in the song (%d)"
        % (len(block_keys)) +
        "; possible corruption")

    if byte_switch_keys == block_keys:
        # No remapping necessary
        return blocks

    new_block_map = {}

    for block_key, byte_switch_key in zip(
            block_keys, byte_switch_keys):

        new_block_map[byte_switch_key] = blocks[block_key]

    return new_block_map


def merge(blocks):
    """Merge the given blocks into a contiguous block of compressed data.

    :param blocks: the list of blocks
    :rtype: a list of compressed bytes
    """

    current_block = blocks[sorted(blocks.keys())[0]]

    compressed_data = []
    eof = False

    while not eof:
        data_size_to_append = None
        next_block = None

        i = 0
        while i < len(current_block.data) - 1:
            current_byte = current_block.data[i]
            next_byte = current_block.data[i + 1]

            if current_byte == RLE_BYTE:
                if next_byte == RLE_BYTE:
                    i += 2
                else:
                    i += 3
            elif current_byte == SPECIAL_BYTE:
                if next_byte in SPECIAL_DEFAULTS:
                    i += 3
                elif next_byte == SPECIAL_BYTE:
                    i += 2
                else:
                    data_size_to_append = i

                    # hit end of file
                    if next_byte == EOF_BYTE:
                        eof = True
                    else:
                        next_block = blocks[next_byte]

                    break
            else:
                i += 1

        assert data_size_to_append is not None, "Ran off the end of a "\
            "block without encountering a block switch or EOF"

        compressed_data.extend(current_block.data[0:data_size_to_append])

        if not eof:
            assert next_block is not None, "Switched blocks, but did " \
                "not provide the next block to switch to"

            current_block = next_block

    return compressed_data


def add_eof(segment):
    """Add an EOF statement to a block."""
    segment.extend([SPECIAL_BYTE, EOF_BYTE])


def add_block_switch(segment, block_id):
    """Add a block switch statement to a block.

    :param segment: the segment to which to add the statement
    :param block_id: the block ID to which the switch statement should switch
    """
    segment.extend([SPECIAL_BYTE, block_id])


def pad(segment, size):
    """Add zeroes to a segment until it reaches a certain size.

    :param segment: the segment to pad
    :param size: the size to which to pad the segment
    """
    for i in range(size - len(segment)):
        segment.append(0)

    assert len(segment) == size


def decompress(compressed_data):
    """Decompress data that has been compressed by the filepack algorithm.

    :param compressed_data: an array of compressed data bytes to decompress

    :rtype: an array of decompressed bytes"""
    raw_data = []

    index = 0

    while index < len(compressed_data):
        current = compressed_data[index]
        index += 1

        if current == RLE_BYTE:
            directive = compressed_data[index]
            index += 1

            if directive == RLE_BYTE:
                raw_data.append(RLE_BYTE)
            else:
                count = compressed_data[index]
                index += 1

                raw_data.extend([directive] * count)
        elif current == SPECIAL_BYTE:
            directive = compressed_data[index]
            index += 1

            if directive == SPECIAL_BYTE:
                raw_data.append(SPECIAL_BYTE)
            elif directive == DEFAULT_WAVE_BYTE:
                count = compressed_data[index]
                index += 1

                raw_data.extend(DEFAULT_WAVE * count)
            elif directive == DEFAULT_INSTR_BYTE:
                count = compressed_data[index]
                index += 1

                raw_data.extend(DEFAULT_INSTRUMENT_FILEPACK * count)
            elif directive == EOF_BYTE:
                assert False, ("Unexpected EOF command encountered while "
                               "decompressing")
            else:
                assert False, "Countered unexpected sequence 0x%02x 0x%02x" % (
                    current, directive)
        else:
            raw_data.append(current)

    return raw_data


def compress(raw_data):
    """Compress raw bytes with the filepack algorithm.

    :param raw_data: an array of raw data bytes to compress

    :rtype: a list of compressed bytes
    """
    raw_data = bytearray(raw_data)
    compressed_data = []

    data_size = len(raw_data)

    index = 0
    next_bytes = [-1, -1, -1]

    def is_default_instrument(index):
        if index + len(DEFAULT_INSTRUMENT_FILEPACK) > len(raw_data):
            return False

        instr_bytes = raw_data[index:index + len(DEFAULT_INSTRUMENT_FILEPACK)]

        if instr_bytes[0] != 0xa8 or instr_bytes[1] != 0:
            return False

        return instr_bytes == DEFAULT_INSTRUMENT_FILEPACK

    def is_default_wave(index):
        return (index + len(DEFAULT_WAVE) <= len(raw_data) and
                raw_data[index:index + len(DEFAULT_WAVE)] == DEFAULT_WAVE)

    while index < data_size:
        current_byte = raw_data[index]

        for i in range(3):
            if index < data_size - (i + 1):
                next_bytes[i] = raw_data[index + (i + 1)]
            else:
                next_bytes[i] = -1

        if current_byte == RLE_BYTE:
            compressed_data.append(RLE_BYTE)
            compressed_data.append(RLE_BYTE)
            index += 1
        elif current_byte == SPECIAL_BYTE:
            compressed_data.append(SPECIAL_BYTE)
            compressed_data.append(SPECIAL_BYTE)
            index += 1
        elif is_default_instrument(index):
            counter = 1
            index += len(DEFAULT_INSTRUMENT_FILEPACK)

            while (is_default_instrument(index) and
                   counter < 0x100):
                counter += 1
                index += len(DEFAULT_INSTRUMENT_FILEPACK)

            compressed_data.append(SPECIAL_BYTE)
            compressed_data.append(DEFAULT_INSTR_BYTE)
            compressed_data.append(counter)

        elif is_default_wave(index):
            counter = 1
            index += len(DEFAULT_WAVE)

            while is_default_wave(index) and counter < 0xff:
                counter += 1
                index += len(DEFAULT_WAVE)

            compressed_data.append(SPECIAL_BYTE)
            compressed_data.append(DEFAULT_WAVE_BYTE)
            compressed_data.append(counter)

        elif (current_byte == next_bytes[0] and
              next_bytes[0] == next_bytes[1] and
              next_bytes[1] == next_bytes[2]):
            # Do RLE compression

            compressed_data.append(RLE_BYTE)
            compressed_data.append(current_byte)
            counter = 0

            while (index < data_size and
                   raw_data[index] == current_byte and
                   counter < 0xff):
                index += 1
                counter += 1

            compressed_data.append(counter)
        else:
            compressed_data.append(current_byte)
            index += 1

    return compressed_data



from .vibrato import Vibrato


class EnvelopeMixin(object):

    @property
    def envelope(self):
        """the noise instrument's volume envelope (8-bit integer)"""
        return self.data.envelope

    @envelope.setter
    def envelope(self, value):
        self.data.envelope = value

    @staticmethod
    def import_lsdinst(obj, struct_data):
        obj.envelope = struct_data['data']['envelope']

    @staticmethod
    def equal(a, b):
        return (isinstance(a, EnvelopeMixin) and isinstance(b, EnvelopeMixin)
                and a.envelope == b.envelope)

class VibratoMixin(object):

    @property
    def vibrato(self):
        """instrument's vibrato settings"""
        return Vibrato(self.data.vibrato)

    @staticmethod
    def import_lsdinst(obj, struct_data):
        Vibrato(obj.data.vibrato).import_lsdinst(struct_data)

    @staticmethod
    def equal(a, b):
        return (isinstance(a, VibratoMixin) and isinstance(b, VibratoMixin)
                and a.vibrato == b.vibrato)



class SoundLengthMixin(object):

    @property
    def sound_length(self):
        """the instrument sound's length, a 6-bit integer or ``unlimited``
        if the sound plays forever"""
        if self.data.has_sound_length:
            return self.data.sound_length
        else:
            return 'unlimited'

    @sound_length.setter
    def sound_length(self, value):
        if value == 'unlimited':
            self.data.sound_length = 0
            self.data.has_sound_length = False
        else:
            self.data.sound_length = value

    @staticmethod
    def import_lsdinst(obj, struct_data):
        if struct_data['data']['has_sound_length']:
            obj.sound_length = struct_data['data']['sound_length']
        else:
            obj.sound_length = 'unlimited'

    @staticmethod
    def equal(a, b):
        return (isinstance(a, SoundLengthMixin)
                and isinstance(b, SoundLengthMixin)
                and a.sound_length == b.sound_length)

class SweepMixin(object):

    @property
    def sweep(self):
        """modulates the sound's frequency; only works on pulse 1
        (8-bit integer)"""
        return self.data.sweep

    @sweep.setter
    def sweep(self, value):
        self.data.sweep = value

    @staticmethod
    def import_lsdinst(obj, struct_data):
        obj.sweep = struct_data['data']['sweep']

    @staticmethod
    def equal(a, b):
        return (isinstance(a, SweepMixin) and isinstance(b, SweepMixin)
                and a.sweep == b.sweep)



class Vibrato(object):

    def __init__(self, data):
        self._data = data

    @property
    def type(self):
        """ ``hf`` (for high frequency sine), ``sawtooth``,
        ``saw`` or ``square``"""
        return self._data.type

    @type.setter
    def type(self, value):
        self._data.type = value

    @property
    def direction(self):
        """'down' or 'up'"""
        return self._data.direction

    @direction.setter
    def direction(self, value):
        self._data.direction = value

    def import_lsdinst(self, struct_data):
        self.direction = struct_data['data']['vibrato']['direction']
        self.type = struct_data['data']['vibrato']['type']

    def __eq__(self, other):
        return (isinstance(other, Vibrato)
                and self.type == other.type
                and self.direction == other.direction)



from .vendor.six.moves import range

# Max. length of a word name
WORD_NAME_LENGTH = 4


def word_cleanup(word):
    return ''.join([chr(x) for x in word if x != 0]).ljust(WORD_NAME_LENGTH)


class Word(object):

    def __init__(self, song, index):
        self._song = song
        """the word's parent song"""

        self._index = index
        """the word's index within the speech instrument"""

    @property
    def song(self):
        return self._song

    @property
    def index(self):
        return self._index

    @property
    def name(self):
        """the word's name"""
        return self._song.song_data.word_names[self._index]

    @name.setter
    def name(self, value):
        self._song.song_data.word_names[self._index] = value

    @property
    def sounds(self):
        """a list of the sounds that make up the word; each sound has
        an ``allophone`` and a ``length``"""
        return self._song.song_data.words[self._index]


class SpeechInstrument(object):

    def __init__(self, song):
        self._song = song

        self._words = [
            Word(self._song, i) for i in
            range(len(self._song.song_data.words))]

    @property
    def song(self):
        """the speech instrument's parent song"""
        return self._song

    @property
    def words(self):
        """a list of the speech instrument's defined words, as Word objects"""
        return self._words



from .instrument import Instrument
from .instrument_mixins import VibratoMixin
from .exceptions import ImportException

from . import bread_spec

from .synth import Synth

from .vendor.six.moves import range

class WaveInstrument(Instrument, VibratoMixin):

    def __init__(self, song, index):
        super(WaveInstrument, self).__init__(song, index)

    @property
    def volume(self):
        """the sound's volume; 0 through 3"""
        return self.data.volume

    @volume.setter
    def volume(self, value):
        self.data.volume = value

    @property
    def synth(self):
        """the wave's synth settings"""
        return self.song.synths[self.data.synth]

    @synth.setter
    def synth(self, value):
        self.data.synth = value.index

    @property
    def repeat(self):
        """the synth sound's repeat point (4-bit integer)"""
        return self.data.repeat

    @repeat.setter
    def repeat(self, value):
        self.data.repeat = value

    @property
    def play_type(self):
        """how to play the synth sound; ``once``, ``loop``, ``ping-pong``,
        or ``manual``"""
        return self.data.play_type

    @play_type.setter
    def play_type(self, value):
        self.data.play_type = value

    @property
    def steps(self):
        """length of the synth sound (4-bit integer)"""
        return self.data.steps

    @steps.setter
    def steps(self, value):
        self.data.steps = value

    @property
    def speed(self):
        """how fast the sound should be played back (4-bit integer)"""
        return self.data.speed

    @speed.setter
    def speed(self, value):
        self.data.speed = value

    def _get_open_synth_index(self):
        available_synths = set(range(bread_spec.NUM_SYNTHS))

        for instrument in self.song.instruments.as_list():
            if instrument is not None and instrument.type == 'wave':
                available_synths.discard(instrument.synth.index)

        if len(available_synths) == 0:
            return None
        else:
            return available_synths.pop()

    def import_lsdinst(self, lsdinst_struct):
        super(WaveInstrument, self).import_lsdinst(lsdinst_struct)

        # Make sure we've got enough space for the synth if we need one
        synth_index = self._get_open_synth_index()

        if synth_index is None:
            raise ImportException(
                "No available synth slot in which to store the instrument's "
                "synth data")

        synth = Synth(self.song, synth_index)
        synth.import_lsdinst(lsdinst_struct['synth'])

        self.synth = synth

        VibratoMixin.import_lsdinst(self, lsdinst_struct)

    def export_struct(self):
        export_struct = super(WaveInstrument, self).export_struct()

        export_struct['synth'] = Synth(self.song, self.synth.index).export()

        return export_struct



import json
import os
import sys
import math
from nose.tools import assert_equal, assert_less

SCRIPT_DIR = os.path.abspath(os.path.dirname(__file__))

sys.path.append(os.path.join(SCRIPT_DIR, os.path.pardir))

from . import filepack as filepack
from .project import Project, load_lsdsng, load_srm
from .utils import temporary_file


def test_save_load_lsdsng():
    sample_song_compressed = os.path.join(
        SCRIPT_DIR, "test_data", "UNTOLDST.lsdsng")

    proj = load_lsdsng(sample_song_compressed)

    expected_song_name = b"UNTOLDST"
    expected_song_version = 23

    assert_equal(proj.name, expected_song_name)
    assert_equal(proj.version, expected_song_version)

    tmp_abspath = None

    with temporary_file() as tmp_abspath:
        proj.save_lsdsng(tmp_abspath)

        read_project = load_lsdsng(tmp_abspath)

        assert_equal(proj, read_project)


def test_read_write_project():
    sample_song_compressed = os.path.join(
        SCRIPT_DIR, "test_data", "sample_song_compressed.json")

    with open(sample_song_compressed, "r") as fp:
        song_data_compressed = json.load(fp)

    song_data = filepack.decompress(song_data_compressed)
    song_name = "UNTOLDST"
    song_version = 23

    # 0xbadf00d for size in blocks is synthetic, since we don't really care
    # about its size for the purposes of this test
    bogus_size_blks = 0xbadf00d

    proj = Project(
        song_name, song_version, bogus_size_blks, song_data)

    assert_equal(proj.name, song_name)
    assert_equal(proj.version, song_version)

    raw_data = proj.get_raw_data()

    recompressed = filepack.compress(raw_data)

    assert_less(math.fabs(len(recompressed) - len(song_data_compressed)), 512)

    # Do comparison based on parsed object, since the actual input data can
    # contain noise
    proj_from_raw_data = Project(
        song_name, song_version, bogus_size_blks, raw_data)

    assert_equal(proj_from_raw_data._song_data, proj._song_data)


def test_block_remap_required():
    block_remap_song = os.path.join(SCRIPT_DIR, "test_data", "ANNARKTE.lsdsng")

    proj = load_lsdsng(block_remap_song)

    assert_equal(b"ANNARKTE", proj.name)
    assert_equal(3, proj.version)
    assert_equal(4, proj.size_blks)


def test_srm_load():
    srm_song = os.path.join(SCRIPT_DIR, "test_data", "sample.srm")

    proj = load_srm(srm_song)
    assert_equal("CLICK", proj.song.instruments[0].name)


def test_save_load_srm():
    srm_song = os.path.join(SCRIPT_DIR, "test_data", "sample.srm")

    proj = load_srm(srm_song)

    with temporary_file() as tmp_proj_abspath:
        proj.save_srm(tmp_proj_abspath)
        read_proj = load_srm(tmp_proj_abspath)
        assert_equal(proj, read_proj)

if __name__ == "__main__":
    test_srm_load()



import os

from nose.tools import assert_equal

from .project import load_lsdsng

SCRIPT_DIR = os.path.abspath(os.path.dirname(__file__))

def test_phrase_load():
    proj = load_lsdsng(os.path.join(SCRIPT_DIR, 'test_data', 'UNTOLDST.lsdsng'))

    phrase = proj.song.phrases[0x24]

    assert_equal('C 4', phrase.notes[0x0])
    assert_equal('K', phrase.fx[0x3])
    assert_equal('L', phrase.fx[0xc])
    assert_equal(1, phrase.fx_val[0xc])
    assert_equal(None, phrase.instruments[1])
    assert_equal(2, phrase.instruments[4].index)



import bread as b

from .vendor.six.moves import range

def padded_hex(pad_count):
    return lambda x: ("0x%%0%dx" % (pad_count)) % (x)

EMPTY_BLOCK = 0xff

# File management structure (starts at 0x8000)
compressed_sav_file = [
    # Up to 32 files (songs) can be saved to one cartridge
    ("filenames", b.array(32, b.string(8))),
    # Each file has a monotonically increasing version number
    ("file_versions", b.array(32, b.byte)),
    b.padding(30 * 8),
    ("sram_init_check", b.string(2)),  # set to 'jk' on init
    # The file that is currently active
    ("active_file", b.byte),
    # Table mapping blocks to files.
    ("block_alloc_table", b.array(191, b.byte))
]

# Should be 0x4000 bytes long
sample_kit = [
    ("magic_number", b.string(2)),  # should be 0x60, 0x40
    # The address of the first byte after each sample, or 0 if the sample is
    # unused
    ("sample_end_addresses", b.array(15, b.string(2))),
    b.padding(2 * 8),
    # For samples with names less than three characters long, sample names are
    # padded with -. Unused sample names are full of nulls.
    ("sample_names", b.array(15, b.string(3))),
    b.padding(8),
    # n/a?
    b.padding(8 * 2),
    ("kit_name", b.string(6)),
    # n/a?
    b.padding(8 * 4),
    # Samples 8-1, then 15-9
    ("force_loop", b.array(15, b.bit)),
    # n/a?
    b.padding(8 * 2),
    # 4 bits per sample at 11468Hz, in chunks of 16 bytes
    ("sample_data", b.array(127, b.array(128, b.nibble)))
]

# Max. number of phrases
NUM_PHRASES = 255

# Max. number of tables
NUM_TABLES = 32

# Number of soft-synths
NUM_SYNTHS = 16

# Waves per soft-synth
WAVES_PER_SYNTH = 16

# Number of frames per wave
FRAMES_PER_WAVE = 32

# Number of entries in each table
ENTRIES_PER_TABLE = 16

# Max. number of instruments
NUM_INSTRUMENTS = 64

# Max. number of sequences in the whole song
NUM_SONG_CHAINS = 256

# Max. number of chains
NUM_CHAINS = 128

# Max. number of phrases per chain
PHRASES_PER_CHAIN = 16

# Max. number of grooves
NUM_GROOVES = 32

# Steps per phrase
STEPS_PER_PHRASE = 16

# Steps per groove
STEPS_PER_GROOVE = 16

# Steps per table
STEPS_PER_TABLE = 16

# Max. length of a "word"
WORD_LENGTH = 0x10

# Number of "words" in the speech instrument
NUM_WORDS = 42

chain = [
    ("pu1", b.byte, {"str_format": padded_hex(2)}),
    ("pu2", b.byte, {"str_format": padded_hex(2)}),
    ("wav", b.byte, {"str_format": padded_hex(2)}),
    ("noi", b.byte, {"str_format": padded_hex(2)})]

pulse_instrument = [
    ("envelope", b.byte, {"str_format": padded_hex(2)}),
    ("phase_transpose", b.byte),
    b.padding(1),
    # If false, sound length is UNLIM
    ("has_sound_length", b.boolean),
    ("sound_length", b.intX(6)),
    ("sweep", b.byte),
    b.padding(3),
    ("automate", b.boolean),
    ("automate_2", b.boolean),
    ("vibrato", [
        ("type", b.enum(2, {
            0: "hf",
            1: "sawtooth",
            2: "sine",
            3: "square"
        })),
        ("direction", b.enum(1, {
            0: "down",
            1: "up"
        }))
    ]),
    b.padding(2),
    ("table_on", b.boolean),
    ("table", b.intX(5)),
    ("wave", b.enum(2, {
        0: "12.5%",
        1: "25%",
        2: "50%",
        3: "75%"
    })),
    ("phase_finetune", b.nibble),
    ("pan", b.enum(2, {
        0: "Invalid",
        1: "L",
        2: "R",
        3: "LR"
    })),
    b.padding(8 * 8)
]

wave_instrument = [
    b.padding(1),
    ("volume", b.enum(2, {
        0: 0,
        1: 3,
        2: 2,
        3: 1
    })),
    b.padding(5),
    ("synth", b.nibble),
    ("repeat", b.nibble),
    b.padding(8 * 2 + 3),
    ("automate", b.boolean),
    ("automate_2", b.boolean),
    ("vibrato", [
        ("type", b.enum(2, {
            0: "hf",
            1: "sawtooth",
            2: "sine",
            3: "square"
        })),
        ("direction", b.enum(1, {
            0: "down",
            1: "up"
        }))
    ]),
    b.padding(2),
    ("table_on", b.boolean),
    ("table", b.intX(5)),
    b.padding(6),
    ("pan", b.enum(2, {
        0: "Invalid",
        1: "L",
        2: "R",
        3: "LR"
    })),
    b.padding(8 + 6),
    ("play_type", b.enum(2, {
        0: "once",
        1: "loop",
        2: "ping-pong",
        3: "manual"
    })),
    b.padding(8 * 4),
    ("steps", b.nibble),
    ("speed", b.nibble, {"str_format": hex, "offset": 1}),
    b.padding(8)
]

kit_instrument = [
    ("volume", b.byte),
    ("keep_attack_1", b.boolean),
    ("half_speed", b.boolean),
    ("kit_1", b.intX(6), {"offset": 1}),
    # 0 == "auto"
    ("length_1", b.byte),
    b.padding(9),
    ("loop_1", b.boolean),
    ("loop_2", b.boolean),
    ("automate", b.boolean),
    ("automate_2", b.boolean),
    ("vibrato", [
        ("type", b.enum(2, {
            0: "hf",
            1: "sawtooth",
            2: "sine",
            3: "square"
        })),
        ("direction", b.enum(1, {
            0: "down",
            1: "up"
        }))
    ]),
    b.padding(2),
    ("table_on", b.boolean),
    ("table", b.intX(5)),
    b.padding(6),
    ("pan", b.enum(2, {
        0: "Invalid",
        1: "L",
        2: "R",
        3: "LR"
    })),
    ("pitch", b.byte),
    ("keep_attack_2", b.boolean),
    b.padding(1),
    ("kit_2", b.intX(6), {"offset": 1}),
    ("dist_type", b.enum(8, {
        0xd0: "clip",
        0xd1: "shape",
        0xd2: "shap2",
        0xd3: "wrap"
    })),
    # 0 == "auto"
    ("length_2", b.byte),
    ("offset_1", b.byte),
    ("offset_2", b.byte),
    b.padding(8 * 2)
]

noise_instrument = [
    ("envelope", b.byte),
    ("s_cmd", b.enum(8, {
        0: "free",
        1: "stable"
    })),
    b.padding(1),
    # If false, sound length is UNLIM
    ("has_sound_length", b.boolean),
    ("sound_length", b.intX(6)),
    ("sweep", b.byte),
    b.padding(3),
    ("automate", b.boolean),
    ("automate_2", b.boolean),
    b.padding(5),
    ("table_on", b.boolean),
    ("table", b.intX(5)),
    b.padding(6),
    ("pan", b.enum(2, {
        0: "Invalid",
        1: "L",
        2: "R",
        3: "LR"
    })),
    b.padding(8 * 8)
]

INSTRUMENT_TYPE_CODE = {
    "pulse": 0,
    "wave": 1,
    "kit": 2,
    "noise": 3
}

instrument = [
    ("instrument_type", b.enum(
        8, dict([(v, k) for (k, v) in list(INSTRUMENT_TYPE_CODE.items())]),
        default="invalid")),
    (b.CONDITIONAL, "instrument_type", {
        "pulse": pulse_instrument,
        "wave": wave_instrument,
        "kit": kit_instrument,
        "noise": noise_instrument,
        "invalid": [b.padding(15 * 8)]
    })
]

softsynth_sound_params = [
    ("volume", b.byte),
    ("filter_cutoff", b.byte),
    ("phase_amount", b.byte),
    ("vertical_shift", b.byte)
]

softsynth = [
    ("waveform", b.enum(8, {
        0: "sawtooth",
        1: "square",
        2: "sine"
    })),
    ("filter_type", b.enum(8, {
        0: "lowpass",
        1: "highpass",
        2: "bandpass",
        3: "allpass"
    })),
    ("filter_resonance", b.byte),
    ("distortion", b.enum(8, {
        0: "clip",
        1: "wrap"
    })),
    ("phase_type", b.enum(8, {
        0: "normal",
        1: "resync",
        2: "resync2"
    })),
    ("start", softsynth_sound_params),
    ("end", softsynth_sound_params),
    b.padding(8 * 3)
]

FX_COMMANDS = {
    0: '-',
    1: 'A',
    2: 'C',
    3: 'D',
    4: 'E',
    5: 'F',
    6: 'G',
    7: 'H',
    8: 'K',
    9: 'L',
    10: 'M',
    11: 'O',
    12: 'P',
    13: 'R',
    14: 'S',
    15: 'T',
    16: 'V',
    17: 'W',
    18: 'Z'
}

table_command = [
    ("fx", b.array(NUM_TABLES,
                   b.array(STEPS_PER_TABLE, b.enum(8, FX_COMMANDS)))),
    ("val", b.array(NUM_TABLES, b.array(STEPS_PER_TABLE, b.byte)))
]

word_sound = [
    ("allophone", b.enum(8, {
        0: '-',
        1: 'AA',
        2: 'AE',
        3: 'AO',
        4: 'AR',
        5: 'AW',
        6: 'AX',
        7: 'AY',
        8: 'BB1',
        9: 'BB2',
        10: 'CH',
        11: 'DD1',
        12: 'DD2',
        13: 'DH1',
        14: 'DH2',
        15: 'EH',
        16: 'EL',
        17: 'ER1',
        18: 'ER2',
        19: 'EY',
        20: 'FF',
        21: 'GG1',
        22: 'GG2',
        23: 'GG3',
        24: 'HH1',
        25: 'HH2',
        26: 'IH',
        27: 'IY',
        28: 'JH',
        29: 'KK1',
        30: 'KK2',
        31: 'KK3',
        32: 'LL',
        33: 'MM',
        34: 'NG',
        35: 'NN1',
        36: 'NN2',
        37: 'OR',
        38: 'OW',
        39: 'OY',
        40: 'PP',
        41: 'RR1',
        42: 'RR2',
        43: 'SH',
        44: 'SS',
        45: 'TH',
        46: 'TT1',
        47: 'TT2',
        48: 'UH',
        49: 'UW1',
        50: 'UW2',
        51: 'VV',
        52: 'WH',
        53: 'WW',
        54: 'XR',
        55: 'YR',
        56: 'YY1',
        57: 'YY2',
        58: 'ZH',
        59: 'ZZ'
    })),
    ("length", b.byte)
]

# A list which provides the names of all the notes store
# in a phrase's `notes` field
NOTES = ['---']

for i in range(0x3, 0x10):
    NOTES.extend(
        ['%s%X' % (x.ljust(2, ' '), i) for x in ('C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#', 'B')])

NOTES_DICT = {}

for i, note in enumerate(NOTES):
    NOTES_DICT[i] = note

song = [
    ("phrase_notes", b.array(NUM_PHRASES, b.array(
        STEPS_PER_PHRASE, b.enum(8, NOTES_DICT)))),
    ("bookmarks", b.array(64, b.byte)),
    b.padding(96 * 8),
    ("grooves", b.array(NUM_GROOVES, b.array(STEPS_PER_GROOVE, b.byte))),
    ("song", b.array(NUM_SONG_CHAINS, chain)),
    ("table_envelopes", b.array(NUM_TABLES, b.array(STEPS_PER_TABLE, b.byte))),
    ("words", b.array(NUM_WORDS, b.array(WORD_LENGTH, word_sound))),
    ("word_names", b.array(NUM_WORDS, b.string(4))),
    # Set to 'rb' on init
    ("mem_init_flag_1", b.string(2)),
    ("instrument_names", b.array(NUM_INSTRUMENTS, b.string(5))),
    b.padding(70 * 8),
    # Beginning of bank 1
    b.padding(32 * 8),
    ("table_alloc_table", b.array(NUM_TABLES, b.byte)),
    ("instr_alloc_table", b.array(NUM_INSTRUMENTS, b.byte)),
    ("chain_phrases", b.array(
        NUM_CHAINS, b.array(PHRASES_PER_CHAIN, b.byte))),
    ("chain_transposes", b.array(
        NUM_CHAINS, b.array(PHRASES_PER_CHAIN, b.byte))),
    ("instruments", b.array(NUM_INSTRUMENTS, instrument)),
    ("table_transposes", b.array(
        NUM_TABLES, b.array(STEPS_PER_TABLE, b.byte))),
    ("table_cmd1", table_command),
    ("table_cmd2", table_command),
    # Set to 'rb' on init
    ("mem_init_flag_2", b.string(2)),
    ("phrase_alloc_table", b.array(NUM_PHRASES, b.boolean)),
    # There are only 255 valid phrases, but the allocation table is 256 bits
    # long, so we ignore the last bit
    b.padding(1),
    ("chain_alloc_table", b.array(NUM_CHAINS, b.boolean)),
    ("softsynth_params", b.array(NUM_SYNTHS, softsynth)),
    ("clock", [
        ("hours", b.byte),
        ("minutes", b.byte)]),
    ("tempo", b.byte),
    ("tune_setting", b.byte),
    ("total_clock", [
        ("days", b.byte),
        ("hours", b.byte),
        ("minutes", b.byte),
        ("checksum", b.byte)
    ]),
    ("key_delay", b.byte),
    ("key_repeat", b.byte),
    ("font", b.byte),
    ("sync_setting", b.enum(8, {
        0: "off",
        1: "slave",
        2: "master",
        3: "midi",
        4: "nano",
        5: "keyboard"
    })),
    ("colorset", b.byte),
    b.padding(8),
    ("clone", b.enum(8, {
        0: "deep",
        1: "slim"
    })),
    ("file_changed", b.byte),
    ("power_save", b.byte),
    ("prelisten", b.byte),
    # One overwrite lock per synth, true if the wave overwrites the synth's
    # parameters; stored in reverse order (synth f .. 0)
    ("wave_synth_overwrite_locks", b.array(NUM_SYNTHS, b.boolean)),
    b.padding(8 * 58),
    # Beginning of bank 2
    ("phrase_fx", b.array(NUM_PHRASES, b.array(
        STEPS_PER_PHRASE, b.enum(8, FX_COMMANDS)))),
    ("phrase_fx_val", b.array(NUM_PHRASES, b.array(STEPS_PER_PHRASE, b.byte))),
    b.padding(32 * 8),
    # Beginning of bank 3
    ("wave_frames",
     b.array(NUM_SYNTHS,
             b.array(WAVES_PER_SYNTH,
                     b.array(FRAMES_PER_WAVE, b.nibble)))),
    ("phrase_instruments",
     b.array(NUM_PHRASES, b.array(STEPS_PER_PHRASE, b.byte))),
    # Set to 'rb' on init
    ("mem_init_flag_3", b.string(2)),
    b.padding(13 * 8),
    ("version", b.byte)
]

# .lsdsng format
lsdsng_preamble = [
    ("name", b.string(8)),
    ("version", b.byte)
]



import math
import bread
from . import bread_spec as spec
from .song import Song
from . import filepack
from . import blockutils
from .blockutils import BlockReader, BlockWriter, BlockFactory


def load_lsdsng(filename):
    """Load a Project from a ``.lsdsng`` file.

    :param filename: the name of the file from which to load
    :rtype: :py:class:`pylsdj.Project`
    """

    # Load preamble data so that we know the name and version of the song
    with open(filename, 'rb') as fp:
        preamble_data = bread.parse(fp, spec.lsdsng_preamble)

    with open(filename, 'rb') as fp:
        # Skip the preamble this time around
        fp.seek(int(len(preamble_data) / 8))

        # Load compressed data into a block map and use BlockReader to
        # decompress it
        factory = BlockFactory()

        while True:
            block_data = bytearray(fp.read(blockutils.BLOCK_SIZE))

            if len(block_data) == 0:
                break

            block = factory.new_block()
            block.data = block_data

        remapped_blocks = filepack.renumber_block_keys(factory.blocks)

        reader = BlockReader()
        compressed_data = reader.read(remapped_blocks)

        # Now, decompress the raw data and use it and the preamble to construct
        # a Project
        raw_data = filepack.decompress(compressed_data)

        name = preamble_data.name
        version = preamble_data.version
        size_blks = int(math.ceil(
            float(len(compressed_data)) / blockutils.BLOCK_SIZE))

        return Project(name, version, size_blks, raw_data)


def load_srm(filename):
    """Load a Project from an ``.srm`` file.

    :param filename: the name of the file from which to load
    :rtype: :py:class:`pylsdj.Project`
    """

    # .srm files are just decompressed projects without headers

    # In order to determine the file's size in compressed blocks, we have to
    # compress it first
    with open(filename, 'rb') as fp:
        raw_data = fp.read()

    compressed_data = filepack.compress(raw_data)

    factory = BlockFactory()
    writer = BlockWriter()
    writer.write(compressed_data, factory)

    size_in_blocks = len(factory.blocks)

    # We'll give the file a dummy name ("SRMLOAD") and version, since we know
    # neither
    name = "SRMLOAD"
    version = 0

    return Project(name, version, size_in_blocks, raw_data)


class Project(object):

    def __init__(self, name, version, size_blks, data):
        self.name = name
        """the project's name"""

        self.version = version
        """the project's version (incremented on every save in LSDJ)"""

        self.size_blks = size_blks
        """the size of the song in filesystem blocks"""

        # Useful for applications tracking whether a project was modified since
        # it was loaded.
        self.modified = False

        # Since parsing the song is expensive, we'll load it lazily from the
        # raw data on-demand
        self.__song_data = None
        self._song = None
        self._raw_bytes = data

    @property
    def _song_data(self):
        if self.__song_data is None:
            self.__song_data = bread.parse(self._raw_bytes, spec.song)

        return self.__song_data

    @_song_data.setter
    def _song_data(self, value):
        self.__song_data = value

    @property
    def song(self):
        """the song associated with the project"""
        if self._song is None:
            self._song = Song(self._song_data)

        return self._song

    @song.setter
    def song(self, value):
        self._song = value

    def get_raw_data(self):
        return bread.write(self._song_data, spec.song)

    def save(self, filename):
        """Save a project in .lsdsng format to the target file.

        :param filename: the name of the file to which to save

        :deprecated: use ``save_lsdsng(filename)`` instead
        """
        with open(filename, 'wb') as fp:
            writer = BlockWriter()
            factory = BlockFactory()

            preamble_dummy_bytes = bytearray([0] * 9)
            preamble = bread.parse(
                preamble_dummy_bytes, spec.lsdsng_preamble)
            preamble.name = self.name
            preamble.version = self.version

            preamble_data = bread.write(preamble)
            raw_data = self.get_raw_data()
            compressed_data = filepack.compress(raw_data)

            writer.write(compressed_data, factory)

            fp.write(preamble_data)

            for key in sorted(factory.blocks.keys()):
                fp.write(bytearray(factory.blocks[key].data))

    def save_lsdsng(self, filename):
        """Save a project in .lsdsng format to the target file.

        :param filename: the name of the file to which to save
        """
        return self.save(filename)

    def save_srm(self, filename):
        """Save a project in .srm format to the target file.

        :param filename: the name of the file to which to save
        """
        with open(filename, 'wb') as fp:
            raw_data = bread.write(self._song_data, spec.song)
            fp.write(raw_data)

    def __eq__(self, other):
        return self._song_data == other._song_data

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return "<%s, %d>\n" % (self.name, self.version)



from setuptools import setup

setup(name='pylsdj',
      version='2.3.2',
      description='A utility belt for dealing with LSDJ-related files',
      url='http://github.com/alexras/pylsdj',
      author='Alex Rasmussen',
      author_email='alexras@acm.org',
      classifiers = [
          'Development Status :: 4 - Beta',
          'Environment :: Other Environment',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Topic :: Software Development :: Libraries :: Python Modules'],
      license='MIT',
      packages=['pylsdj', 'pylsdj.vendor'],
      requires=['bread'],
      install_requires=['bread>=2.1.0'],
      zip_safe=False)



from . import filepack

"""LSDJ stores its data in a block-oriented format. This file contains a Block
object that encapsulates a block's data, and
"""

# Maximum size of a block - 512 bytes
BLOCK_SIZE = 0x200


class Block(object):

    def __init__(self, block_id, data):
        self.id = block_id
        self.data = data

    def __repr__(self):  # pragma: no cover
        repr_str = "{"
        repr_str += "BLOCK %d (%d bytes)\n" % (self.id, len(self.data))

        i = 0

        while i < len(self.data):
            repr_str += '\t'.join(
                ["0x%02x" % (x) for x in
                 self.data[i:min(i + 10, len(self.data))]]) + "\n"

            i += 10

        repr_str += "}"
        return repr_str


class BlockFactory(object):

    """Each block's ID should correspond to its position in an array of blocks.
    """

    def __init__(self):
        self.max_id = 0
        self.blocks = {}

    def new_block(self):
        block = Block(self.max_id, [])
        self.blocks[self.max_id] = block
        self.max_id += 1

        return block


class BlockWriter(object):

    def write(self, compressed_data, factory):
        """Splits a compressed byte stream into blocks.
        """
        return filepack.split(compressed_data, BLOCK_SIZE,
                              factory)


class BlockReader(object):

    def read(self, block_dict):
        """Parses a dictionary of blocks into a compressed byte stream.
        """

        return filepack.merge(block_dict)



from .utils import assert_index_sane
import json


class Instrument(object):

    def __init__(self, song, index):
        self.song = song
        self.data = song.song_data.instruments[index]
        self.index = index

    def __eq__(self, other):
        return (isinstance(other, Instrument) and
                self.name == other.name and
                self.type == other.type and
                self.table == other.table and
                self.automate == other.automate and
                self.pan == other.pan)

    @property
    def name(self):
        """the instrument's name (5 characters, zero-padded)"""
        instr_name = self.song.song_data.instrument_names[self.index]

        if type(instr_name) == bytes:
            instr_name = instr_name.decode('utf-8')

        return instr_name

    @name.setter
    def name(self, val):
        if type(val) != bytes:
            val = val.encode('utf-8')

        self.song.song_data.instrument_names[self.index] = val

    @property
    def type(self):
        """the instrument's type (``pulse``, ``wave``, ``kit`` or ``noise``)"""
        return self.data.instrument_type

    @type.setter
    def type(self, value):
        self.data.instrument_type = value

    @property
    def table(self):
        """a ```pylsdj.Table``` referencing the instrument's table, or None
        if the instrument doesn't have a table"""
        if hasattr(self.data, 'table_on') and self.data.table_on:
            assert_index_sane(self.data.table, len(self.song.tables))
            return self.song.tables[self.data.table]

    @table.setter
    def table(self, value):
        if not hasattr(self.data, "table_on"):
            raise ValueError("This instrument doesn't have a table")

        self.data.table_on = True
        self.data.table = value.index

    @property
    def automate(self):
        """if True, automation is on"""
        return self.data.automate

    @automate.setter
    def automate(self, value):
        self.data.automate = value

    @property
    def pan(self):
        return self.data.pan

    @pan.setter
    def pan(self, value):
        self.data.pan = value

    def import_lsdinst(self, struct_data):
        """import from an lsdinst struct"""
        self.name = struct_data['name']
        self.automate = struct_data['data']['automate']
        self.pan = struct_data['data']['pan']

        if self.table is not None:
            self.table.import_lsdinst(struct_data)

    def export_to_file(self, filename):
        """Export this instrument's settings to a file.

        :param filename: the name of the file
        """
        instr_json = self.export_struct()

        with open(filename, 'w') as fp:
            json.dump(instr_json, fp, indent=2)

    def export_struct(self):
        export_struct = {}

        export_struct['name'] = self.name
        export_struct['data'] = {}

        data_json = json.loads(self.data.as_json())

        for key, value in list(data_json.items()):
            if key[0] != '_' and key not in ('synth', 'table'):
                export_struct['data'][key] = value

        if self.table is not None:
            export_struct['table'] = self.table.export()

        return export_struct



from .instrument import Instrument
from .instrument_mixins import EnvelopeMixin, SoundLengthMixin, VibratoMixin, \
    SweepMixin

MIXINS = [VibratoMixin, SoundLengthMixin, EnvelopeMixin, SweepMixin]


class PulseInstrument(Instrument,
                      VibratoMixin, SoundLengthMixin, EnvelopeMixin,
                      SweepMixin):

    def __init__(self, song, index):
        super(PulseInstrument, self).__init__(song, index)

    def __eq__(self, other):
        return (super(PulseInstrument, self).__eq__(other) and
                isinstance(other, PulseInstrument) and
                self.phase_transpose == other.phase_transpose and
                self.phase_finetune == other.phase_finetune and
                self.wave == other.wave and
                VibratoMixin.equal(self, other) and
                SoundLengthMixin.equal(self, other) and
                EnvelopeMixin.equal(self, other) and
                SweepMixin.equal(self, other))

    @property
    def phase_transpose(self):
        """detune pulse channel 2 this many semitones; in LSDJ, this is
        ``PU2 TUNE`` (8-bit integer)"""
        return self.data.phase_transpose

    @phase_transpose.setter
    def phase_transpose(self, value):
        self.data.phase_transpose = value

    @property
    def phase_finetune(self):
        """detune pulse channel 1 down, channel 2 up; in LSDJ, this is
        ``PU FINE`` (4-bit integer)"""
        return self.data.phase_finetune

    @phase_finetune.setter
    def phase_finetune(self, value):
        self.data.phase_finetune = value

    @property
    def wave(self):
        """the pulse's wave width; ``12.5%``, ``25%``, ``50%`` or ``75%``"""
        return self.data.wave

    @wave.setter
    def wave(self, value):
        self.data.wave = value

    def import_lsdinst(self, struct_data):
        super(PulseInstrument, self).import_lsdinst(struct_data)

        self.phase_transpose = struct_data['data']['phase_transpose']
        self.phase_finetune = struct_data['data']['phase_finetune']
        self.wave = struct_data['data']['wave']

        for mixin in MIXINS:
            mixin.import_lsdinst(self, struct_data)
